import os
import sqlite3

from pathlib import Path

import numpy as np

from nanocompore.eventalign_collapse import EventalignCollapser
from nanocompore.common import MEASUREMENTS_TYPE

from tests.common import cwd


def test_eventalign_collapse():
    """
    An integration tests that runs the whole
    eventalign_collapse process.
    """
    tsv_in = os.path.join(cwd, 'tests/fixtures/kd1_eventalign.tsv')
    ref = os.path.join(cwd, 'tests/fixtures/test_reference.fa')
    db_out = 'collapsed.sqlite'

    collapser = EventalignCollapser(tsv_in, ref, db_out, 2, '.')

    try:
        collapser()

        conn = sqlite3.connect(db_out)
        num_transcripts = conn.execute('SELECT COUNT(*) FROM transcripts').fetchone()[0]
        assert num_transcripts == 2

        signal_data = conn.execute('SELECT * FROM signal_data').fetchall()

        # Validate that we have data for two transcripts
        assert {1, 2} == {row[0] for row in signal_data}

        # Validate that we have four reads
        read_ids = [row[1] for row in signal_data]
        assert len(read_ids) == 4

        query = f'''SELECT read
                    FROM reads
                    WHERE id IN ({",".join(["?" for _ in range(len(read_ids))])})'''
        read_uuids = conn.execute(query, read_ids).fetchall()
        read_uuids = {row[0] for row in read_uuids}

        assert read_uuids == {'3f46f499-8ce4-4817-8177-8ad61b784f27',
                              '73d62df4-f04a-4207-a4bc-7b9739b3c3b2',
                              'b7bc9a36-318e-4be2-a90f-74a5aa6439bf',
                              '65db1ced-cbf5-40e4-880b-9f202c715804'}

        # Validate the measurements for the first two
        # positions of the first read.
        intensities = np.frombuffer(signal_data[0][2], dtype=MEASUREMENTS_TYPE)
        assert round(float(intensities[301]), 3) == 117.218
        assert round(float(intensities[302]), 3) == 71.906

        dwells = np.frombuffer(signal_data[0][3], dtype=MEASUREMENTS_TYPE)
        assert round(float(dwells[301]), 3) == 0.008
        assert round(float(dwells[302]), 3) == 0.003

        # Validate the calculation of read invalid
        # kmer ratio.
        reads_ratios = {read: invalid_ratio
                        for read, _, invalid_ratio in conn.execute('SELECT * FROM reads').fetchall()}
        # Read 3f46f499-8ce4-4817-8177-8ad61b784f27
        # is for transcript:
        # ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|
        # We have three positions in the eventalign tsv
        # with positions 302 and 303 being valid.
        # Position 301 has two events, one of which is
        # NNNNN, so the position is invalid.
        # The expected invalid ratio should be
        # the (alignment_length - num_valid)/alignment_length
        assert reads_ratios['3f46f499-8ce4-4817-8177-8ad61b784f27'] == (3 - 2)/3
    finally:
        # Make sure to delete the sqlite database
        Path(db_out).unlink()


def test_read_data():
    tsv_in = os.path.join(cwd, 'tests/fixtures/kd1_eventalign.tsv')
    ref = os.path.join(cwd, 'tests/fixtures/test_reference.fa')
    db_out = 'collapsed.sqlite'

    collapser = EventalignCollapser(tsv_in, ref, db_out, 2, '.')

    fstart = 161
    fend = 6649

    with open(tsv_in) as f:
        data = collapser._read_data(f, fstart, fend)

    read, kmers = data[0]

    assert read == '3f46f499-8ce4-4817-8177-8ad61b784f27'
    assert len(kmers) == 3
    assert kmers[0].pos == 301

    # Test that the two events for the position
    # are collapsed:
    assert kmers[0].dwell == 0.00498 + 0.00332
    # The median intensity should be the median
    # of all sample values from the two events.
    assert kmers[0].median_intensity == 117.218
    # The first event is valid, but the second
    # is not, so the collapsed kmer should be
    # invalid as well.
    assert not kmers[0].valid


def test_process_ref():
    tsv_in = os.path.join(cwd, 'tests/fixtures/kd1_eventalign.tsv')
    ref = os.path.join(cwd, 'tests/fixtures/test_reference.fa')
    db_out = 'collapsed.sqlite'

    collapser = EventalignCollapser(tsv_in, ref, db_out, 2, '.')

    fstart = 161
    fend = 6649

    with open(tsv_in) as f:
        data = collapser._read_data(f, fstart, fend)

    ref_id = 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|'
    ref_len = 2554

    intensity, dwell, read_invalid_ratios = collapser._process_ref(ref_id, data, ref_len)

    assert intensity.shape == (3, 2554)
    assert round(float(intensity[0, 301]), 3) == 117.218

    assert dwell.shape == (3, 2554)
    assert round(float(dwell[0, 301]), 4) == 0.00498 + 0.00332

    assert len(read_invalid_ratios) == 3
    assert round(read_invalid_ratios['3f46f499-8ce4-4817-8177-8ad61b784f27'], 3) == 0.333




#!/usr/bin/env python3
import argparse
import os
import shutil
import textwrap
import yaml
import sys

from loguru import logger

from nanocompore import __version__ as package_version
from nanocompore import __description__ as package_description
from nanocompore.common import Kit
from nanocompore.common import NanocomporeError
from nanocompore.common import mkdir
from nanocompore.config import Config
from nanocompore.eventalign_collapse import EventalignCollapser
from nanocompore.preprocessing import RemoraPreprocessor
from nanocompore import plotting
from nanocompore.run import RunCmd


def main(args=None):
    # General parser
    parser = argparse.ArgumentParser(description=package_description, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version', '-v', action='version', version='v'+package_version)

    subparser_description = textwrap.dedent(f"""
            nanocompore implements the following subcommands
            \t● {'template':<20} : Initialize a new input configuration file using the default template.
            \t● {'eventalign_collapse':<20} : Parse eventalign data to process and store it to an intermediary
                                           efficient SQLite database for later analysis.
            \t● {'remora_resquiggle':<20} : Use Remora to resquiggle a sample and create an intermediary
                                           efficient SQLite database for later analysis.
            \t● {'run':<20} : Compare 2 samples and find significant signal differences.
            \t● {'plot':<20} : Generate plots.""")
    subparsers = parser.add_subparsers(dest='subcommand',
                                       required=True,
                                       description=subparser_description)

    # Sampcomp subparser
    parser_sc = subparsers.add_parser('run',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      description=textwrap.dedent("Compare 2 samples and find significant signal differences."))
    parser_sc.add_argument('config', type=str)
    parser_sc.set_defaults(func=sampcomp_subcommand)

    # preprocess eventalign_collapse
    parser_sc = subparsers.add_parser('eventalign_collapse',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      description=textwrap.dedent("Parse eventalign data to process and store it to an intermediary efficient database for later analysis."))
    parser_sc.add_argument('--ref', '-r', help="Transcriptome fasta reference.", required=True)
    parser_sc.add_argument('--input', '-i', help="Path to input eventalign file. If not provided, the input is read from stdin (useful for piping nanopolish/f5c eventalign directly).")
    parser_sc.add_argument('--output', '-o', help="Path to output SQLite database.", required=True)
    parser_sc.add_argument('--nthreads', '-n', help="Number of parallel processes to use for processing.", nargs='?', type=int, const=2, default=2)
    parser_sc.add_argument('--tmp', '-t', help="Directory where tmp files would be created (default: current directory).", nargs='?', type=str, const='.', default='.')
    parser_sc.set_defaults(func=eventalign_collapse_subcommand)

    # preprocess remora_resquiggle
    parser_sc = subparsers.add_parser('remora_resquiggle',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      description=textwrap.dedent("Use Remora to resquiggle a sample and create an SQLite DB with the signal measurements."))
    parser_sc.add_argument('--ref', '-r', help="Transcriptome fasta reference.", required=True)
    parser_sc.add_argument('--pod5', '-p', help="Path to input pod5 file containing the raw signal data.", required=True)
    parser_sc.add_argument('--bam', '-b', help="Path to input bam file containing the aligned reads.", required=True)
    parser_sc.add_argument('--output', '-o', help="Path to output SQLite database.", required=True)
    parser_sc.add_argument('--kit', '-k', help="Sequencing kit that was use (should be RNA002 or RNA004).", required=True)
    parser_sc.add_argument('--max-reads', '-m', help="Maximum number of reads to resquiggle for a transcript (default: 5000).", nargs='?', type=int, const=5000, default=5000)
    parser_sc.add_argument('--nthreads', '-n', help="Number of parallel processes to use for processing (default: 2).", nargs='?', type=int, const=2, default=2)
    parser_sc.set_defaults(func=remora_resquiggle_subcommand)

    # Init subparser
    parser_init = subparsers.add_parser('template',
                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                        description=textwrap.dedent("Initialize a new input configuration file using the default template."))
    parser_init.add_argument('--overwrite', '-o', action='store_true', help="Overwrite existing config file.")
    parser_init.add_argument('path', type=str)
    parser_init.set_defaults(func=init_subcommand)

    parser_plot = subparsers.add_parser('plot',
                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                        description=textwrap.dedent("Plotting functionality."))
    # plot_subparsers = parser_plot.add_subparsers(help='Plot type.')
    plot_subparsers_description = textwrap.dedent(f"""
            nanocompore plot implements the following subcommands
            \t● {'pvalues':<10} : Plot the p-values for a reference region.
            \t● {'signal':<10} : Plot the signal measurements for a region.
                                 Note: this will plot the data from the input
                                 files without applying any filtering performed
                                 by Nanocompore.
            \t● {'position':<10} : Plot the signal data for a specific position as
                                 a 2D plot. Note: this will plot the data as it is
                                 in the input files, without applying the filtering
                                 performed by Nanocompore.
            \t● {'gmm':<10} : Plot the GMM fitting for a position.
                                 This replicates all data filtering and produces
                                 identical GMM fitting to the one obtained by the
                                 run command.
            \t● {'coverage':<10} : Plot the read coverage over a reference region.
                                 Note: this will plot the data as it is in the input
                                 files, without applying the filtering performed
                                 by Nanocompore.""")
    plot_subparsers = parser_plot.add_subparsers(dest='subcommand',
                                                 required=True,
                                                 description=plot_subparsers_description)

    parser_pvalues = plot_subparsers.add_parser("pvalues",
                                                formatter_class=argparse.RawDescriptionHelpFormatter,
                                                description=textwrap.dedent("Plot the p-values for a reference region."))
    parser_pvalues.add_argument('--config', '-c', type=str, required=True, help="Path to the input configuration YAML file.")
    parser_pvalues.add_argument('--output', '-o', type=str, required=True, help="Path and filename where the plot will be saved.")
    parser_pvalues.add_argument('reference', type=str, help="Reference name, matching a name from the FASTA reference used in the config.")
    parser_pvalues.add_argument('--start', type=int, help="0-based index on the reference.")
    parser_pvalues.add_argument('--end', type=int, help="0-based index on the reference.")
    parser_pvalues.add_argument('--threshold', type=float, help="Threshold p-value that will be drawn as a line.", default=0.01)
    parser_pvalues.add_argument('--kind', type=str, help="Kind of plot to draw. Default: lineplot", choices=['lineplot', 'barplot'], default='lineplot')
    parser_pvalues.add_argument('--figsize', type=str, help="Figure size. Default: 30,10", default='30,10')
    parser_pvalues.add_argument('--tests', type=str, help="Which tests to plot. By default all executed tests are shown.", default=None)
    parser_pvalues.add_argument('--palette', type=str, help="Color palette to use. Default: Dark2", default='Dark2')
    parser_pvalues.set_defaults(func=plot_pvalues)

    parser_signal = plot_subparsers.add_parser("signal",
                                               formatter_class=argparse.RawDescriptionHelpFormatter,
                                               description=textwrap.dedent("Plot the signal measurements for a region."))
    parser_signal.add_argument('--config', '-c', type=str, required=True, help="Path to the input configuration YAML file.")
    parser_signal.add_argument('--output', '-o', type=str, required=True, help="Path and filename where the plot will be saved.")
    parser_signal.add_argument('reference', type=str, help="Reference name, matching a name from the FASTA reference used in the config.")
    parser_signal.add_argument('--start', type=int, help="0-based index on the reference.")
    parser_signal.add_argument('--end', type=int, help="0-based index on the reference.")
    parser_signal.add_argument('--kind', type=str, help="Kind of plot to draw. Default: violinplot", choices=['violinplot', 'swarmplot', 'boxenplot'], default='violinplot')
    parser_signal.add_argument('--figsize', type=str, help="Figure size: WIDTH,HEIGHT. Default: 30,10", default='30,10')
    parser_signal.add_argument('--split_samples', action='store_true', help="Split results by sample instead of by condition.")
    parser_signal.add_argument('--markersize', type=int, help="Size of the points if swarmplot is used.", default=2)
    parser_signal.add_argument('--palette', type=str, help="Color palette to use. Default: Dark2", default='Dark2')
    parser_signal.set_defaults(func=plot_signal)

    parser_position = plot_subparsers.add_parser("position",
                                                 formatter_class=argparse.RawDescriptionHelpFormatter,
                                                 description=textwrap.dedent("Plot the signal data for a specific position as a 2D plot."))
    parser_position.add_argument('--config', '-c', type=str, required=True, help="Path to the input configuration YAML file.")
    parser_position.add_argument('--output', '-o', type=str, required=True, help="Path and filename where the plot will be saved.")
    parser_position.add_argument('reference', type=str, help="Reference name, matching a name from the FASTA reference used in the config.")
    parser_position.add_argument('position', type=int, help="Position to plot (0-based index).")
    parser_position.add_argument('--figsize', type=str, help="Figure size: WIDTH,HEIGHT. Default: 10,10", default='10,10')
    parser_position.add_argument('--point_size', type=int, help="Size of the data points", default=20)
    parser_position.add_argument('--xlim', type=str, help="Set specific range for the x-axis: MIN,MAX. By default it will be inferred from the data")
    parser_position.add_argument('--ylim', type=str, help="Set specific range for the x-axis: MIN,MAX. By default it will be inferred from the data")
    parser_position.add_argument('--kde', action='store_true', help="Plot the KDE of the intensity/dwell bivarariate distributions in the two samples.")
    parser_position.add_argument('--kde_levels', type=int, help="How many levels of the gaussian distributions to show", default=10)
    parser_position.add_argument('--palette', type=str, help="Use a single palette to select both point and gmm colors.", default='Dark2')
    parser_position.set_defaults(func=plot_position)

    parser_gmm = plot_subparsers.add_parser("gmm",
                                            formatter_class=argparse.RawDescriptionHelpFormatter,
                                            description=textwrap.dedent("Plot the GMM fitting for a single position."))
    parser_gmm.add_argument('--config', '-c', type=str, required=True, help="Path to the input configuration YAML file.")
    parser_gmm.add_argument('--output', '-o', type=str, required=True, help="Path and filename where the plot will be saved.")
    parser_gmm.add_argument('reference', type=str, help="Reference name, matching a name from the FASTA reference used in the config.")
    parser_gmm.add_argument('position', type=int, help="Position to plot (0-based index).")
    parser_gmm.add_argument('--figsize', type=str, help="Figure size: WIDTH,HEIGHT. Default: 10,10", default='10,10')
    parser_gmm.add_argument('--point_size', type=int, help="Size of the data points", default=20)
    parser_gmm.add_argument('--xlim', type=str, help="Set specific range for the x-axis: MIN,MAX. By default it will be inferred from the data")
    parser_gmm.add_argument('--ylim', type=str, help="Set specific range for the x-axis: MIN,MAX. By default it will be inferred from the data")
    parser_gmm.add_argument('--gmm_levels', type=int, help="How many levels of the gaussian distributions to show", default=4)
    parser_gmm.add_argument('--palette', type=str, help="Use a single palette to select both point and gmm colors.", default='Dark2')
    parser_gmm.add_argument('--point_palette', type=str, help="Which palette to use for plotting the points.")
    parser_gmm.add_argument('--gmm_palette', type=str, help="Which palette to use for plotting the gaussian distributions.")
    parser_gmm.set_defaults(func=plot_gmm)

    parser_coverage = plot_subparsers.add_parser("coverage",
                                                 formatter_class=argparse.RawDescriptionHelpFormatter,
                                                 description=textwrap.dedent("Plot the read coverage for a region."))
    parser_coverage.add_argument('--config', '-c', type=str, required=True, help="Path to the input configuration YAML file.")
    parser_coverage.add_argument('--output', '-o', type=str, required=True, help="Path and filename where the plot will be saved.")
    parser_coverage.add_argument('reference', type=str, help="Reference name, matching a name from the FASTA reference used in the config.")
    parser_coverage.add_argument('--start', type=int, help="0-based index on the reference.")
    parser_coverage.add_argument('--end', type=int, help="0-based index on the reference.")
    parser_coverage.add_argument('--figsize', type=str, help="Figure size: WIDTH,HEIGHT. Default: 30,10", default='12,4')
    parser_coverage.add_argument('--split_samples', action='store_true', help="Split results by sample instead of by condition.")
    parser_coverage.add_argument('--palette', type=str, help="Use a single palette to select both point and gmm colors.", default='Dark2')
    parser_coverage.set_defaults(func=plot_coverage)

    if len(sys.argv) == 1:
        # no argument passed
        args = parser.parse_args(['--help'])
    elif len(sys.argv) == 2 and sys.argv[1] == 'plot':
        # plot called without arguments
        args = parser.parse_args(['plot', '--help'])
    else:
        args = parser.parse_args()

    # Call relevant subcommand function
    args.func(args)


#~~~~~~~~~~~~~~SUBCOMMAND FUNCTIONS~~~~~~~~~~~~~~#


def sampcomp_subcommand(args):
    """
    Runs the sample comparison subcommand.
    """

    # Read the input config file
    with open(args.config, 'r') as f:
        config_file = yaml.safe_load(f)

    try:
        config = Config(config_file)
    except Exception as e:
        msg = f"ERROR: {e}\n"
        sys.stderr.write(msg)
        exit(1)

    # If we'll show a progress bar then we
    # want to prevent the default loguru
    # logging on stderr.
    if config.get_progress:
        logger.remove()

    # Check if output folder already exists
    try:
        mkdir(fn=config.get_outpath(), exist_ok=config.get_result_exists_strategy() != 'stop')
    except (NanocomporeError, FileExistsError) as E:
        raise NanocomporeError("Could not create the output folder. Try using `overwrite: true` in the input configuration or use another directory")

    setup_logger(config, "run.log")

    run_cmd = RunCmd(config)
    run_cmd()


def eventalign_collapse_subcommand(args):
    """
    Parse eventalign data to process and store it
    to an intermediary efficient database for later
    analysis.
    """
    EventalignCollapser(args.input, args.ref, args.output, args.nthreads, args.tmp)()


def remora_resquiggle_subcommand(args):
    """
    Resquiggle a sample with Remora.
    """
    kit = Kit[args.kit]
    RemoraPreprocessor(args.ref,
                       args.pod5,
                       args.bam,
                       args.output,
                       kit,
                       args.max_reads,
                       args.nthreads)()


def init_subcommand(args):
    """
    Initializes a new input configuration file using the default template.
    """
    template = os.path.join(os.path.dirname(__file__), 'template_config.yaml')
    if os.path.isfile(args.path) and not args.overwrite:
        raise NanocomporeError("Output file already exists. Use --overwrite to overwrite it.")
    shutil.copyfile(template, args.path)


def plot_pvalues(args: argparse.Namespace):
    """
    Plot pvalues along a reference region.

    Parameters
    ----------
    args : argparse.Namespace
        Command-line arguments
    """
    # Read the input config file
    with open(args.config, 'r') as f:
        config_file = yaml.safe_load(f)

    try:
        config = Config(config_file)
    except Exception as e:
        msg = f"ERROR: {e}\n"
        sys.stderr.write(msg)
        exit(1)

    figsize = tuple(map(int, args.figsize.split(',')))
    tests = args.tests.split(',') if args.tests else None

    fig = plotting.plot_pvalues(config,
                                args.reference,
                                args.start,
                                args.end,
                                args.kind,
                                args.threshold,
                                figsize,
                                tests,
                                args.palette)
    fig.savefig(args.output)


def plot_signal(args: argparse.Namespace):
    """
    Plot the signal measurements along a reference region.

    Parameters
    ----------
    args : argparse.Namespace
        Command-line arguments
    """
    # Read the input config file
    with open(args.config, 'r') as f:
        config_file = yaml.safe_load(f)

    try:
        config = Config(config_file)
    except Exception as e:
        msg = f"ERROR: {e}\n"
        sys.stderr.write(msg)
        exit(1)

    figsize = tuple(map(int, args.figsize.split(',')))

    fig = plotting.plot_signal(config,
                               args.reference,
                               args.start,
                               args.end,
                               args.kind,
                               figsize,
                               args.split_samples,
                               args.markersize,
                               args.palette)
    fig.savefig(args.output)


def plot_position(args: argparse.Namespace):
    """
    Plot the signal measurements for a specific position.

    Parameters
    ----------
    args : argparse.Namespace
        Command-line arguments
    """
    # Read the input config file
    with open(args.config, 'r') as f:
        config_file = yaml.safe_load(f)

    try:
        config = Config(config_file)
    except Exception as e:
        msg = f"ERROR: {e}\n"
        sys.stderr.write(msg)
        exit(1)

    figsize = tuple(map(int, args.figsize.split(',')))
    xlim = None
    if args.xlim is not None:
        xlim = tuple(map(int, args.xlim.split(',')))
    ylim = None
    if args.ylim is not None:
        ylim = tuple(map(int, args.ylim.split(',')))

    fig = plotting.plot_position(config,
                                 args.reference,
                                 args.position,
                                 figsize,
                                 args.point_size,
                                 xlim,
                                 ylim,
                                 args.kde,
                                 args.kde_levels,
                                 args.palette)
    fig.savefig(args.output)

def plot_gmm(args: argparse.Namespace):
    """
    Plot GMM for a position.

    Parameters
    ----------
    args : argparse.Namespace
        Command-line arguments
    """
    # Read the input config file
    with open(args.config, 'r') as f:
        config_file = yaml.safe_load(f)

    try:
        config = Config(config_file)
    except Exception as e:
        msg = f"ERROR: {e}\n"
        sys.stderr.write(msg)
        exit(1)

    figsize = tuple(map(int, args.figsize.split(',')))
    xlim = None
    if args.xlim is not None:
        xlim = tuple(map(int, args.xlim.split(',')))
    ylim = None
    if args.ylim is not None:
        ylim = tuple(map(int, args.ylim.split(',')))

    fig = plotting.plot_gmm(config,
                            args.reference,
                            args.position,
                            figsize,
                            args.point_size,
                            xlim,
                            ylim,
                            args.gmm_levels,
                            args.palette,
                            args.point_palette,
                            args.gmm_palette)
    fig.savefig(args.output)


def plot_coverage(args: argparse.Namespace):
    """
    Plot the read coverage for a region.

    Parameters
    ----------
    args : argparse.Namespace
        Command-line arguments
    """
    # Read the input config file
    with open(args.config, 'r') as f:
        config_file = yaml.safe_load(f)

    try:
        config = Config(config_file)
    except Exception as e:
        msg = f"ERROR: {e}\n"
        sys.stderr.write(msg)
        exit(1)

    figsize = tuple(map(int, args.figsize.split(',')))

    fig = plotting.plot_coverage(config,
                                 args.reference,
                                 args.start,
                                 args.end,
                                 figsize,
                                 args.split_samples,
                                 args.palette)
    fig.savefig(args.output)


def setup_logger(config, file_name):
    if config.get_result_exists_strategy() == 'continue':
        logger_mode = 'a'
    else:
        logger_mode = 'w'
    logger.add(os.path.join(config.get_outpath(),
                            file_name),
                mode=logger_mode,
                level=config.get_log_level())


#~~~~~~~~~~~~~~CLI ENTRYPOINT~~~~~~~~~~~~~~#

if __name__ == "__main__":
    main()




import numpy as np

from scipy.stats import kstest
from scipy.stats import multivariate_normal
from sklearn.preprocessing import StandardScaler


def gof_test_singlerep(data, conditions, config):
    control_label = config.get_depleted_condition()
    control_label_id = config.get_condition_ids()[control_label]

    control_data = data[conditions == control_label_id, :]
    test_data = data[conditions != control_label_id, :]

    scaler = StandardScaler()
    x_control = scaler.fit_transform(control_data)
    x_control = x_control[~np.isnan(x_control)]
    x_test = scaler.transform(test_data)
    x_test = x_test[~np.isnan(x_test)]

    mean_control = np.mean(x_control, axis=0)
    cov_control = np.cov(x_control, rowvar=0)

    probs_control = multivariate_normal.pdf(x_control,
                                            mean=mean_control,
                                            cov=cov_control,
                                            allow_singular=True)
    probs_test = multivariate_normal.pdf(x_test,
                                         mean=mean_control,
                                         cov=cov_control,
                                         allow_singular=True)

    result = kstest(probs_test, probs_control)
    return result.pvalue


def gof_test_multirep(data, samples, conditions, config):
    control_label = config.get_depleted_condition()
    test_label = config.get_test_condition()

    internal_likelihoods = []
    external_likelihoods = []

    control_samples = config.get_condition_samples(control_label)
    test_samples = config.get_condition_samples(test_label)

    for sample in control_samples:
        sample_data = data[samples == sample]
        sample_mean = np.mean(sample_data, axis=0)
        sample_cov = np.cov(sample_data, rowvar=False)

        tech_replicates = [s for s in control_samples if s != sample]
        for replicate in tech_replicates:
            replicate_data = data[samples == replicate]
            likelihood = multivariate_normal.pdf(replicate_data,
                                                 mean=sample_mean,
                                                 cov=sample_cov,
                                                 allow_singular=True)
            internal_likelihoods.append(likelihood)

        for bio_replicate in test_samples:
            replicate_data = data[samples == bio_replicate]
            likelihood = multivariate_normal.pdf(replicate_data,
                                                 mean=sample_mean,
                                                 cov=sample_cov,
                                                 allow_singular=True)
            external_likelihoods.append(likelihood)

    internal_likelihoods = np.concatenate(internal_likelihoods)
    external_likelihoods = np.concatenate(external_likelihoods)

    result = kstest(external_likelihoods, internal_likelihoods)
    return result.pvalue




import sys

from unittest.mock import patch

import pytest

from nanocompore.__main__ import main


def test_main_no_args_shows_help(capsys):
    with pytest.raises(SystemExit) as exception:
        with patch.object(sys, 'argv', ['nanocompore']):
            main()

    assert exception.type == SystemExit
    assert exception.value.code == 0
    captured = capsys.readouterr()
    assert "nanocompore implements the following subcommands" in captured.out




class Transcript():
    def __init__(self, id, ref_id, ref_seq):
        self._id = id
        self._name = ref_id
        self._ref_seq = ref_seq
        self._length = len(self._ref_seq)


    @property
    def id(self):
        return self._id


    @property
    def name(self):
        return self._name


    @property
    def length(self):
        return self._length


    @property
    def seq(self):
        return self._ref_seq




import copy

from unittest.mock import Mock, call

import torch
import numpy as np
import pandas as pd

from nanocompore.comparisons import TranscriptComparator
from nanocompore.comparisons import calculate_lors
from nanocompore.comparisons import get_contigency_matrices
from nanocompore.config import Config
from nanocompore.common import MOTOR_DWELL_POS

from tests.common import BASIC_CONFIG
from tests.common import MockWorker
from tests.common import naneq


def test_add_shift_stats():
    config = Config(BASIC_CONFIG)
    data = torch.tensor([
        # pos 0
        [
            # The columns are:
            # intensity, dwell
            [1.3, 1.1], # read 0
            [1.2, 0.1], # read 1
            [1.3, 0.3], # read 2
            [1.0, 0.2], # read 3
        ],
        # pos 1
        [
            [1.3, 1.1], # read 4
            [1.2, 0.1], # read 5
            [1.1, 0.2], # read 6
            [0.6, 0.4], # read 6
        ],
    ])
    conditions = torch.tensor([0, 1, 0, 1])


    comparator = TranscriptComparator(config, MockWorker())

    results = {}
    comparator._add_shift_stats(results,
                                data,
                                conditions,
                                'cpu')

    assert get_float(results['c1_mean_intensity'][0]) == 1.3
    assert get_float(results['c1_mean_intensity'][1]) == 1.2
    assert get_float(results['c1_mean_dwell'][0]) == 0.7
    assert get_float(results['c1_mean_dwell'][1]) == 0.65

    assert get_float(results['c2_mean_intensity'][0]) == 1.1
    assert get_float(results['c2_mean_intensity'][1]) == 0.9
    assert get_float(results['c2_mean_dwell'][0]) == 0.15
    assert get_float(results['c2_mean_dwell'][1]) == 0.25

    assert get_float(results['c1_sd_intensity'][1]) == 0.1


def test_add_shift_stats_with_motor_dwell():
    config_yaml = copy.deepcopy(BASIC_CONFIG)
    config_yaml['motor_dwell_offset'] = 11
    config = Config(config_yaml)
    data = torch.tensor([
        # pos 0
        [
            # The columns are:
            # intensity, dwell, motor_dwell
            [1.3, 1.1, 0.6], # read 0
            [1.2, 0.1, 1.0], # read 1
            [1.3, 0.3, 0.4], # read 2
            [1.0, 0.2, 0.3], # read 3
        ],
        # pos 1
        [
            [1.3, 1.1, 0.6], # read 4
            [1.2, 0.1, 1.0], # read 5
            [1.1, 0.2, 0.4], # read 6
            [0.6, 0.4, 0.2], # read 6
        ],
    ])
    conditions = torch.tensor([0, 1, 0, 1])

    comparator = TranscriptComparator(config, MockWorker())

    results = {}
    comparator._add_shift_stats(results,
                                data,
                                conditions,
                                'cpu')

    assert get_float(results['c1_mean_intensity'][0]) == 1.3
    assert get_float(results['c1_mean_intensity'][1]) == 1.2
    assert get_float(results['c1_mean_dwell'][0]) == 0.7
    assert get_float(results['c1_mean_dwell'][1]) == 0.65
    assert get_float(results['c1_mean_motor_dwell'][0]) == 0.5
    assert get_float(results['c1_mean_motor_dwell'][1]) == 0.5

    assert get_float(results['c2_mean_intensity'][0]) == 1.1
    assert get_float(results['c2_mean_intensity'][1]) == 0.9
    assert get_float(results['c2_mean_dwell'][0]) == 0.15
    assert get_float(results['c2_mean_dwell'][1]) == 0.25
    assert get_float(results['c2_mean_motor_dwell'][0]) == 0.65
    assert get_float(results['c2_mean_motor_dwell'][1]) == 0.6

    assert get_float(results['c1_sd_intensity'][1]) == 0.1


def test_nonparametric_test_KS():
    config = Config(BASIC_CONFIG)
    comparator = TranscriptComparator(config, MockWorker())
    data = torch.tensor([
        # pos 3
        [
            [97.5, np.log10(0.08), np.log10(0.3)], # read 1
            [98.8, np.log10(0.012), np.log10(0.03)], # read 2
            [84.3, np.log10(0.4), np.nan], # read 3
            [113.2, np.log10(0.06), np.nan], # read 4
            [np.nan, np.nan, np.nan], # read 5
            [np.nan, np.nan, np.nan] # read 6
            ],
        # pos 5
        [
            [98.5, np.log10(0.09), np.nan], # read 1
            [99.8, np.log10(0.013), np.nan], # read 2
            [85.3, np.log10(0.5), np.nan], # read 3
            [114.2, np.log10(0.07), np.nan], # read 4
            [np.nan, np.nan, np.nan], # read 5
            [np.nan, np.nan, np.nan] # read 6
            ],
        # pos 14 - motor kmer of pos 3
        [
            [82.3, np.log10(0.3), np.nan], # read 1
            [111.2, np.log10(0.03), np.nan], # read 2
            [np.nan, np.nan, np.nan], # read 3
            [np.nan, np.nan, np.nan], # read 4
            [96.5, np.log10(0.07), np.nan], # read 5
            [97.8, np.log10(0.01), np.nan] # read 6
            ]
        ], dtype=float)
    conditions = torch.tensor([0, 1, 0, 1, 0, 1])

    results = comparator._nonparametric_test('KS', data, conditions)

    assert 'KS_intensity_pvalue' in results
    assert 'KS_dwell_pvalue' in results

    assert len(results['KS_intensity_pvalue']) == 3
    assert len(results['KS_dwell_pvalue']) == 3


def test_combine_context_pvalues():
    yaml = copy.deepcopy(BASIC_CONFIG)
    yaml['sequence_context_weights'] = 'harmonic'
    yaml['sequence_context'] = 3
    config = Config(yaml)
    comparator = TranscriptComparator(config, MockWorker())
    results = pd.DataFrame({'pos': range(1, 17),
                            'GMM_chi2_pvalue': [0.001,
                                                np.nan,
                                                0.01,
                                                0.103,
                                                0.1,
                                                0.301,
                                                0.901,
                                                np.nan,
                                                0.001,
                                                0.701,
                                                np.nan,
                                                0.98,
                                                0.99,
                                                0.03,
                                                0.99,
                                                0.99]})
    # Remove the first two nan to simulate low coverage
    # positions that were not tested. The third nan is
    # left in place to simulate a tested position that
    # has failed.
    results.drop([1, 7], axis='index', inplace=True)

    actual = comparator._combine_context_pvalues(results, 'GMM_chi2_pvalue')
    actual = np.array(actual)

    # These values are the results taken from the
    # v1 implementation of the context merging.
    expected = np.array([0.002123656975661474,
                         0.0019018653049881877,
                         0.00711410035724805,
                         0.06601202462390741,
                         0.04439107441454254,
                         0.17963717597756168,
                         0.011383510423954581,
                         0.27956447545269386,
                         np.nan,
                         0.471918623078623,
                         0.8109385225797305,
                         0.3482236872871895,
                         0.8342987523763911,
                         0.9504729672694502])
    assert np.equal(np.nan_to_num(actual.round(3), -1),
                    np.nan_to_num(expected.round(3), -1)).all()


def test_calculate_lor():
    contingencies = torch.tensor([
        # pos 0
        [[13, 11],
         [12, 10]],
        # pos 1
        [[9, 11],
         [7, 21]],
        # pos 2
        [[2, 1],
         [7, 1]]])
    assert torch.all(calculate_lors(contingencies) == torch.tensor([-0.015, 0.898, -1.253]))


def test_get_contigency_matrices():
    # conditions has shape (reads,)
    conditions = torch.tensor([0, 1, 0, 1, 0, 1])
    # predictions has shape (positions, reads)
    predictions = torch.tensor([[0, 0, 0, 1, 1, 1],
                                [0, 1, 0, 1, 0, 1]])
    expected = torch.tensor([
        # pos 0 contengency matrix
        [[2.0, 1],
         [1, 2]],
        # pos 1 contingency matrix
        [[3, 0],
         [0, 3]]])
    assert torch.all(get_contigency_matrices(conditions, predictions) == expected)


def test_get_contigency_matrices_with_gaps():
    # conditions has shape (reads,)
    conditions = torch.tensor([0, 1, 0, 1, 0, 1])
    # predictions has shape (positions, reads)
    predictions = torch.tensor([[0, 0, 0, np.nan, 1, 1],
                                [0, 1, 0, np.nan, 0, 1]])
    expected = torch.tensor([
        # pos 0 contengency matrix
        [[2.0, 1],
         [1, 1]],
        # pos 1 contingency matrix
        [[3, 0],
         [0, 2]]])
    assert torch.all(get_contigency_matrices(conditions, predictions) == expected)


def test_get_cluster_counts():
    config = Config(BASIC_CONFIG)
    # Condition 0 is the knockdown (depleted condition).
    # E.g. for the first position:
    # Most of the points for the non-depleted condition (1)
    # are found in cluster 1, then we can guess that the
    # modification cluster is 1.
    # For the third, cluster 0 will be the modified one.
    conditions = torch.tensor([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    samples = torch.tensor([0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 3])
    # We have predictions for three positions.
    predictions = torch.tensor([[0, 0, 1, 0, 0, 1, 0, 0, 1, 1, 1],
                                [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1],
                                [1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0]])

    contingency = get_contigency_matrices(conditions, predictions)

    comparator = TranscriptComparator(config, MockWorker())

    cluster_counts = comparator._get_cluster_counts(contingency, samples, conditions, predictions)

    assert np.all(cluster_counts['kd1_mod'] == np.array([0, 0, 0]))
    assert np.all(cluster_counts['kd2_mod'] == np.array([1, 0, 0]))
    assert np.all(cluster_counts['wt1_mod'] == np.array([1, 2, 2]))
    assert np.all(cluster_counts['wt2_mod'] == np.array([3, 4, 4]))
    assert np.all(cluster_counts['kd1_unmod'] == np.array([2, 2, 2]))
    assert np.all(cluster_counts['kd2_unmod'] == np.array([2, 3, 3]))
    assert np.all(cluster_counts['wt1_unmod'] == np.array([1, 0, 0]))
    assert np.all(cluster_counts['wt2_unmod'] == np.array([1, 0, 0]))


def test_gmm_test():
    config_yaml = copy.deepcopy(BASIC_CONFIG)
    config_yaml['motor_dwell_offset'] = 1
    config = Config(config_yaml)
    comparator = TranscriptComparator(config, MockWorker())

    data = torch.tensor([
        # pos 0 (2d test because high proportion of missing motor dwell)
        [[1., 1., np.nan],
         [np.nan, np.nan, 1.0],
         [1., 1., np.nan],
         [np.nan, np.nan, 1.0],
         [1., 1., 1.0],
         [1., 1., np.nan]],
        # pos 1 (should be run with 3d test)
        [[2., 2., 2.0],
         [2., 2., 2.0],
         [2., 2., 2.0],
         [2.0, 2.0, 2.0],
         [1., 1., 2.0],
         [2., 2., 2.0]],
        # pos 2 (2d test because high proportion of missing motor dwell)
        [[np.nan, np.nan, np.nan],
         [np.nan, np.nan, 3.0],
         [3., 3., np.nan],
         [np.nan, np.nan, 3.0],
         [3., 3., np.nan],
         [np.nan, np.nan, np.nan]],
    ])
    samples = torch.arange(6)
    conditions = torch.tensor([0, 0, 0, 1, 1, 1])

    gmm_results_2d = {'GMM_chi2_pvalue': np.array([0.1, 0.82]),
                      'GMM_LOR': np.array([0.8, 0.2])}
    gmm_results_3d = {'GMM_chi2_pvalue': np.array([0.01]),
                      'GMM_LOR': np.array([1.3])}

    def test(data, samples, conditions, device):
        if data.shape[2] == 3:
            return gmm_results_3d
        return gmm_results_2d

    comparator._gmm_test_split = Mock(side_effect=test)

    results = comparator._gmm_test(data, samples, conditions, 'cpu')

    callargs = comparator._gmm_test_split.call_args_list

    # Validate that the gmm_test_split is called once for pos 1 with 3d data
    # and once for positions 0 and2 with 2d data. Since we don't want to
    # impose the order we try both ways.
    assert ((naneq(callargs[0].args[0], data[[True, False, True], :, :MOTOR_DWELL_POS])
             and
             naneq(callargs[1].args[0], data[[False, True, False]]))
            or
            (naneq(callargs[0].args[0], data[[False, True, False]])
             and
             naneq(callargs[1].args[0], data[[True, False, True], :, :MOTOR_DWELL_POS])))

    # Validate that the results of the 2d and 3d tests
    # are properly merged in the correct order.
    assert np.array_equal(results['GMM_chi2_pvalue'], np.array([0.1, 0.01, 0.82]))
    assert np.array_equal(results['GMM_LOR'], np.array([0.8, 1.3, 0.2]))


def test_gmm_test_split_single_component():
    config = Config(BASIC_CONFIG)
    comparator = TranscriptComparator(config, MockWorker())

    rand_gen = np.random.default_rng(seed=42)
    data = torch.tensor(rand_gen.standard_normal((1, 100, 2)))
    samples = torch.tensor([samp
                            for samp in [0, 1, 2, 3]
                            for _ in range(25)])
    conditions = torch.tensor([cond
                               for cond in [0, 1]
                               for _ in range(50)])

    # The data is sampled from a normal distribution
    # and one component should fit it better (i.e.
    # should have lower BIC value). We should've
    # detected that and omit reporting the p-value
    # from the 2-component GMM.
    results = comparator._gmm_test_split(data, samples, conditions, 'cpu')
    assert len(results['GMM_chi2_pvalue']) == 1
    assert np.isnan(results['GMM_chi2_pvalue'][0])


def test_gmm_test_split_two_components():
    config = Config(BASIC_CONFIG)
    comparator = TranscriptComparator(config, MockWorker())

    rand_gen = np.random.default_rng(seed=42)
    data = torch.tensor(rand_gen.standard_normal((1, 100, 2)))
    data[:, 50:, :] += 3

    samples = torch.tensor([samp
                            for samp in [0, 1, 2, 3]
                            for _ in range(25)])
    conditions = torch.tensor([cond
                               for cond in [0, 1]
                               for _ in range(50)])

    results = comparator._gmm_test_split(data, samples, conditions, 'cpu')
    assert len(results['GMM_chi2_pvalue']) == 1
    assert results['GMM_chi2_pvalue'] < 0.01


def get_float(value):
    return round(float(value), 3)




import copy
import os
import tempfile
import shutil

import numpy as np
import pandas as pd

from nanocompore.postprocessing import Postprocessor
from nanocompore.config import Config

from tests.common import BASIC_CONFIG


def test_multipletests_filter_nan():
    pvals = np.array([0.1, 0.01, np.nan, 0.01, 0.5, 0.4, 0.01, 0.001, np.nan, np.nan, 0.01, np.nan])
    expected_qvals = np.array([0.13333333, 0.016, np.nan, 0.016, 0.5, 0.45714286, 0.016, 0.008, np.nan, np.nan, 0.016, np.nan])

    qvals = Postprocessor._multipletests_filter_nan(pvals)

    assert np.allclose(qvals, expected_qvals, equal_nan=True)


def test_export_results_tsv_single_chunk_no_genomic_positions():
    config_yaml = copy.deepcopy(BASIC_CONFIG)

    # Since we may run tests in parallel we want to
    # avoid having two tests write at the same path
    # simultaneously.
    with tempfile.TemporaryDirectory() as tmp:
        config_yaml['outpath'] = tmp

        shutil.copyfile('tests/fixtures/out_sampComp_sql.db',
                        os.path.join(tmp, 'out_sampComp_sql.db'))

        results_path = os.path.join(tmp, 'out_nanocompore_results.tsv')

        config = Config(config_yaml)
        postprocessor = Postprocessor(config)

        all_columns = postprocessor._db.get_result_column_names()
        test_result_columns = postprocessor._get_test_result_columns(all_columns)
        postprocessor._export_results_tsv(test_result_columns)

        results = pd.read_csv(results_path, sep='\t')

        assert results.shape == (3437, 21)
        assert set(results.ref_id.unique()) == {'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|',
                                                'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|'}
        expected_cols = ['pos',
                         'chr',
                         'genomicPos',
                         'ref_id',
                         'strand',
                         'ref_kmer',
                         'GMM_chi2_pvalue',
                         'GMM_chi2_qvalue',
                         'KS_dwell_pvalue',
                         'KS_dwell_qvalue',
                         'KS_intensity_pvalue',
                         'KS_intensity_qvalue',
                         'GMM_LOR',
                         'KD1_mod',
                         'KD1_unmod',
                         'KD2_mod',
                         'KD2_unmod',
                         'WT1_mod',
                         'WT1_unmod',
                         'WT2_mod',
                         'WT2_unmod']
        assert np.all(results.columns == expected_cols)
        assert np.all(np.isnan(results.genomicPos))
        assert np.all(np.isnan(results.chr))
        assert np.all(np.isnan(results.strand))


def test_export_results_tsv_single_chunk_with_genomic_positions():
    config_yaml = copy.deepcopy(BASIC_CONFIG)
    config_yaml['gtf'] = 'tests/fixtures/test_annotation.gtf'

    # Since we may run tests in parallel we want to
    # avoid having two tests write at the same path
    # simultaneously.
    with tempfile.TemporaryDirectory() as tmp:
        config_yaml['outpath'] = tmp

        shutil.copyfile('tests/fixtures/out_sampComp_sql.db',
                        os.path.join(tmp, 'out_sampComp_sql.db'))

        results_path = os.path.join(tmp, 'out_nanocompore_results.tsv')

        config = Config(config_yaml)
        postprocessor = Postprocessor(config)

        all_columns = postprocessor._db.get_result_column_names()
        test_result_columns = postprocessor._get_test_result_columns(all_columns)
        postprocessor._export_results_tsv(test_result_columns)

        results = pd.read_csv(results_path, sep='\t')

        assert results.shape == (3437, 21)
        assert set(results.ref_id.unique()) == {'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|',
                                                'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|'}
        expected_cols = ['pos',
                         'chr',
                         'genomicPos',
                         'ref_id',
                         'strand',
                         'ref_kmer',
                         'GMM_chi2_pvalue',
                         'GMM_chi2_qvalue',
                         'KS_dwell_pvalue',
                         'KS_dwell_qvalue',
                         'KS_intensity_pvalue',
                         'KS_intensity_qvalue',
                         'GMM_LOR',
                         'KD1_mod',
                         'KD1_unmod',
                         'KD2_mod',
                         'KD2_unmod',
                         'WT1_mod',
                         'WT1_unmod',
                         'WT2_mod',
                         'WT2_unmod']
        assert np.all(results.columns == expected_cols)

        test_position = results[(results.pos == 89) & (results.ref_id == 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|')].iloc[0]
        assert test_position.genomicPos == 5529654
        assert test_position.chr == 'chr7'
        assert test_position.strand == '-'

        test_position = results[(results.pos == 1982) & (results.ref_id == 'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|')].iloc[0]
        assert test_position.genomicPos == 5527184
        assert test_position.chr == 'chr7'
        assert test_position.strand == '-'


def test_export_results_tsv_multiple_chunks_no_genomic_positions():
    config_yaml = copy.deepcopy(BASIC_CONFIG)

    # Since we may run tests in parallel we want to
    # avoid having two tests write at the same path
    # simultaneously.
    with tempfile.TemporaryDirectory() as tmp:
        config_yaml['outpath'] = tmp

        shutil.copyfile('tests/fixtures/out_sampComp_sql.db',
                        os.path.join(tmp, 'out_sampComp_sql.db'))

        results_path = os.path.join(tmp, 'out_nanocompore_results.tsv')

        config = Config(config_yaml)
        postprocessor = Postprocessor(config)

        all_columns = postprocessor._db.get_result_column_names()
        test_result_columns = postprocessor._get_test_result_columns(all_columns)
        postprocessor._export_results_tsv(test_result_columns, chunksize=100)

        results = pd.read_csv(results_path, sep='\t')

        assert results.shape == (3437, 21)
        assert set(results.ref_id.unique()) == {'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|',
                                                'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|'}
        expected_cols = ['pos',
                         'chr',
                         'genomicPos',
                         'ref_id',
                         'strand',
                         'ref_kmer',
                         'GMM_chi2_pvalue',
                         'GMM_chi2_qvalue',
                         'KS_dwell_pvalue',
                         'KS_dwell_qvalue',
                         'KS_intensity_pvalue',
                         'KS_intensity_qvalue',
                         'GMM_LOR',
                         'KD1_mod',
                         'KD1_unmod',
                         'KD2_mod',
                         'KD2_unmod',
                         'WT1_mod',
                         'WT1_unmod',
                         'WT2_mod',
                         'WT2_unmod']
        assert np.all(results.columns == expected_cols)
        assert np.all(np.isnan(results.genomicPos))
        assert np.all(np.isnan(results.chr))
        assert np.all(np.isnan(results.strand))


def test_export_results_tsv_multple_chunks_with_genomic_positions():
    config_yaml = copy.deepcopy(BASIC_CONFIG)
    config_yaml['gtf'] = 'tests/fixtures/test_annotation.gtf'

    # Since we may run tests in parallel we want to
    # avoid having two tests write at the same path
    # simultaneously.
    with tempfile.TemporaryDirectory() as tmp:
        config_yaml['outpath'] = tmp

        shutil.copyfile('tests/fixtures/out_sampComp_sql.db',
                        os.path.join(tmp, 'out_sampComp_sql.db'))

        results_path = os.path.join(tmp, 'out_nanocompore_results.tsv')

        config = Config(config_yaml)
        postprocessor = Postprocessor(config)

        all_columns = postprocessor._db.get_result_column_names()
        test_result_columns = postprocessor._get_test_result_columns(all_columns)
        postprocessor._export_results_tsv(test_result_columns, chunksize=100)

        results = pd.read_csv(results_path, sep='\t')

        assert results.shape == (3437, 21)
        assert set(results.ref_id.unique()) == {'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|',
                                                'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|'}
        expected_cols = ['pos',
                         'chr',
                         'genomicPos',
                         'ref_id',
                         'strand',
                         'ref_kmer',
                         'GMM_chi2_pvalue',
                         'GMM_chi2_qvalue',
                         'KS_dwell_pvalue',
                         'KS_dwell_qvalue',
                         'KS_intensity_pvalue',
                         'KS_intensity_qvalue',
                         'GMM_LOR',
                         'KD1_mod',
                         'KD1_unmod',
                         'KD2_mod',
                         'KD2_unmod',
                         'WT1_mod',
                         'WT1_unmod',
                         'WT2_mod',
                         'WT2_unmod']
        assert np.all(results.columns == expected_cols)

        test_position = results[(results.pos == 89) & (results.ref_id == 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|')].iloc[0]
        assert test_position.genomicPos == 5529654
        assert test_position.chr == 'chr7'
        assert test_position.strand == '-'

        test_position = results[(results.pos == 1982) & (results.ref_id == 'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|')].iloc[0]
        assert test_position.genomicPos == 5527184
        assert test_position.chr == 'chr7'
        assert test_position.strand == '-'


def test_export_shift_stats_single_chunk():
    config_yaml = copy.deepcopy(BASIC_CONFIG)

    # Since we may run tests in parallel we want to
    # avoid having two tests write at the same path
    # simultaneously.
    with tempfile.TemporaryDirectory() as tmp:
        config_yaml['outpath'] = tmp

        shutil.copyfile('tests/fixtures/out_sampComp_sql.db',
                        os.path.join(tmp, 'out_sampComp_sql.db'))

        results_path = os.path.join(tmp, 'out_nanocompore_shift_stats.tsv')

        config = Config(config_yaml)
        postprocessor = Postprocessor(config)

        shift_columns = postprocessor._get_shift_stats_columns()
        postprocessor._export_shift_stats(shift_columns)

        results = pd.read_csv(results_path, sep='\t')

        assert results.shape == (3437, 14)
        assert set(results.ref_id.unique()) == {'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|',
                                                'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|'}
        expected_cols = ['ref_id',
                         'pos',
                         'c1_mean_intensity',
                         'c2_mean_intensity',
                         'c1_median_intensity',
                         'c2_median_intensity',
                         'c1_sd_intensity',
                         'c2_sd_intensity',
                         'c1_mean_dwell',
                         'c2_mean_dwell',
                         'c1_median_dwell',
                         'c2_median_dwell',
                         'c1_sd_dwell',
                         'c2_sd_dwell']
        assert np.all(results.columns == expected_cols)

        test_position = results[(results.pos == 89) & (results.ref_id == 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|')].iloc[0]
        assert test_position.c1_mean_intensity == 12116.667
        assert test_position.c1_mean_dwell == 1.7454

        test_position = results[(results.pos == 1982) & (results.ref_id == 'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|')].iloc[0]
        assert test_position.c1_mean_intensity == 7186.5
        assert test_position.c1_mean_dwell == 1.3476


def test_export_shift_stats_multiple_chunks():
    config_yaml = copy.deepcopy(BASIC_CONFIG)

    # Since we may run tests in parallel we want to
    # avoid having two tests write at the same path
    # simultaneously.
    with tempfile.TemporaryDirectory() as tmp:
        config_yaml['outpath'] = tmp

        shutil.copyfile('tests/fixtures/out_sampComp_sql.db',
                        os.path.join(tmp, 'out_sampComp_sql.db'))

        results_path = os.path.join(tmp, 'out_nanocompore_shift_stats.tsv')

        config = Config(config_yaml)
        postprocessor = Postprocessor(config)

        shift_columns = postprocessor._get_shift_stats_columns()
        postprocessor._export_shift_stats(shift_columns, chunksize=100)

        results = pd.read_csv(results_path, sep='\t')

        assert results.shape == (3437, 14)
        assert set(results.ref_id.unique()) == {'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|',
                                                'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|'}
        expected_cols = ['ref_id',
                         'pos',
                         'c1_mean_intensity',
                         'c2_mean_intensity',
                         'c1_median_intensity',
                         'c2_median_intensity',
                         'c1_sd_intensity',
                         'c2_sd_intensity',
                         'c1_mean_dwell',
                         'c2_mean_dwell',
                         'c1_median_dwell',
                         'c2_median_dwell',
                         'c1_sd_dwell',
                         'c2_sd_dwell']
        assert np.all(results.columns == expected_cols)

        test_position = results[(results.pos == 89) & (results.ref_id == 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|')].iloc[0]
        assert test_position.c1_mean_intensity == 12116.667
        assert test_position.c1_mean_dwell == 1.7454

        test_position = results[(results.pos == 1982) & (results.ref_id == 'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|')].iloc[0]
        assert test_position.c1_mean_intensity == 7186.5
        assert test_position.c1_mean_dwell == 1.3476




import copy
import multiprocessing
import os
import shutil
import tempfile

import numpy as np
import torch

from pyfaidx import Fasta

from nanocompore.common import INTENSITY_POS
from nanocompore.common import DWELL_POS
from nanocompore.common import MOTOR_DWELL_POS
from nanocompore.config import Config
from nanocompore.run import RunCmd
from nanocompore.run import Worker
from nanocompore.run import Uncalled4Worker
from nanocompore.run import GenericWorker
from nanocompore.transcript import Transcript
from tests.common import BASIC_CONFIG
from tests.common import BASIC_EVENTALIGN_CONFIG
from tests.common import naneq


def test_run():
    """
    A very simple integration test that just
    checks that the run command can finish
    successfully and produces the correct
    number of rows for the fixture bams.
    """
    config_yaml = copy.deepcopy(BASIC_CONFIG)
    output_path = tempfile.mkdtemp(prefix="test_")
    config_yaml['outpath'] = output_path
    config_yaml['min_coverage'] = 1

    # This is a workaround for a very nasty issue
    # where running this integration test after
    # test_comparisons would cause it to crash with
    # a segmentation fault.
    # I've traced the problem and it appears that when
    # we run the GMM in the following tests:
    # - test_comparisons.test_gmm_test_split_single_component
    # - test_comparisons.test_gmm_test_split_two_components
    # then initialising any tensors in the run flow here
    # causes the segfault in the GOMP library.
    # Probably we're hitting some incompatibility
    # between the multiprocessing and the MPI used
    # by pytorch.
    # The workaround is to use spawn() to start the
    # worker processes, so that the torch memory
    # is not shared between the processes.
    # We then revert the change back to using
    # fork().
    multiprocessing.set_start_method("spawn", force=True)
    try:
        config = Config(config_yaml)
        run_cmd = RunCmd(config)

        run_cmd()

        with open(os.path.join(output_path, "out_nanocompore_results.tsv")) as f:
            assert len(f.readlines()) == 3520
    finally:
        shutil.rmtree(output_path)
        multiprocessing.set_start_method("fork", force=True)


def test_prepare_data_no_motor():
    config = Config(BASIC_CONFIG)

    worker = Worker(1, None, None, None, 0, 'cpu', config)

    data = np.array([
        # pos 0
        [[1., 1.],
         [1., 1.]],
        # pos 1
        [[2., 2.],
         [2., 2.]],
        # pos 2
        [[3., 3.],
         [3., 3.]],
    ])
    sample_ids = np.arange(3)
    condition_ids = np.array([0, 0, 1])

    tensor, samples, conditions, positions = worker._prepare_data(data, sample_ids, condition_ids)

    assert isinstance(tensor, torch.Tensor)
    assert tensor.shape == (3, 2, 2)
    assert tensor[0, 0, INTENSITY_POS] == 1.
    # The dwell is log transformed (and we add an epsilon
    # ot avoid division by zero when the value is 0).
    assert tensor[0, 0, DWELL_POS] == np.log10(1.0 + 1e-10)

    assert isinstance(samples, torch.Tensor)
    assert torch.equal(samples, torch.tensor([0, 1, 2]))

    assert isinstance(conditions, torch.Tensor)
    assert torch.equal(conditions, torch.tensor([0, 0, 1]))

    assert isinstance(positions, torch.Tensor)
    assert torch.equal(positions, torch.tensor([0, 1, 2]))


def test_prepare_data_with_motor():
    config_yaml = copy.deepcopy(BASIC_CONFIG)
    config_yaml['motor_dwell_offset'] = 2
    config = Config(config_yaml)

    worker = Worker(1, None, None, None, 0, 'cpu', config)

    data = np.array([
        # pos 0
        [[1., 1.],
         [1., 1.]],
        # pos 1
        [[2., 2.],
         [2., 2.]],
        # pos 2
        [[3., 3.],
         [3., 3.]],
    ])
    sample_ids = np.arange(3)
    condition_ids = np.array([0, 0, 1])

    tensor, samples, conditions, positions = worker._prepare_data(data, sample_ids, condition_ids)

    assert isinstance(tensor, torch.Tensor)
    assert tensor.shape == (3, 2, 3)
    assert tensor[0, 0, INTENSITY_POS] == 1.
    # The dwell is log transformed (and we add an epsilon
    # ot avoid division by zero when the value is 0).
    assert tensor[0, 0, DWELL_POS] == np.log10(1.0 + 1e-10)
    # The first position gets the motor from the 3rd,
    # and the remaining two positions have no motor dwell.
    assert tensor[0, 0, MOTOR_DWELL_POS] == np.log10(3.0 + 1e-10)
    assert tensor[1, 0, MOTOR_DWELL_POS].isnan()
    assert tensor[2, 0, MOTOR_DWELL_POS].isnan()

    assert isinstance(samples, torch.Tensor)
    assert torch.equal(samples, torch.tensor([0, 1, 2]))

    assert isinstance(conditions, torch.Tensor)
    assert torch.equal(conditions, torch.tensor([0, 0, 1]))

    assert isinstance(positions, torch.Tensor)
    assert torch.equal(positions, torch.tensor([0, 1, 2]))


def test_downsample():
    config = Config(BASIC_CONFIG)

    worker = Worker(1, None, None, None, 0, 'cpu', config)

    # covered positions per read are respectively:
    # 2, 1, 3, 0, 3, 2
    data = torch.tensor([
        # pos 0
        [[1., 1.],
         [np.nan, np.nan],
         [1., 1.],
         [np.nan, np.nan],
         [1., 1.],
         [1., 1.]],
        # pos 1
        [[2., 2.],
         [2., 2.],
         [2., 2.],
         [np.nan, np.nan],
         [1., 1.],
         [2., 2.]],
        # pos 2
        [[np.nan, np.nan],
         [np.nan, np.nan],
         [3., 3.],
         [np.nan, np.nan],
         [3., 3.],
         [np.nan, np.nan]],
    ])

    sample_ids = torch.arange(6)
    condition_ids = torch.tensor([0, 0, 0, 1, 1, 1])

    max_reads = 2
    r = worker._downsample(data, sample_ids, condition_ids, max_reads)
    data, samples, conditions = r

    assert data.shape == (3, 4, 2)
    assert torch.equal(samples, torch.tensor([0, 2, 4, 5]))
    assert torch.equal(conditions, torch.tensor([0, 0, 1, 1]))


def test_filter_processed_transcripts():
    config = Config(BASIC_CONFIG)

    worker = Worker(1, None, None, None, 0, 'cpu', config)

    # coverage per position is respectively:
    # cond 0
    # 2, 3, 1
    # cond 1
    # 2, 2, 1
    data = torch.tensor([
        # pos 0
        [[1., 1.],
         [np.nan, np.nan],
         [1., 1.],
         [np.nan, np.nan],
         [1., 1.],
         [1., 1.]],
        # pos 1
        [[2., 2.],
         [2., 2.],
         [2., 2.],
         [np.nan, np.nan],
         [1., 1.],
         [2., 2.]],
        # pos 2
        [[np.nan, np.nan],
         [np.nan, np.nan],
         [3., 3.],
         [np.nan, np.nan],
         [3., 3.],
         [np.nan, np.nan]],
    ])

    positions = torch.arange(3)
    conditions = torch.tensor([0, 0, 0, 1, 1, 1])

    min_cov = 2
    filtered_data, positions = worker._filter_low_cov_positions(data, positions, conditions, min_cov)

    assert filtered_data.shape == (2, 6, 2)
    assert naneq(filtered_data, data[:2])
    assert torch.equal(positions, torch.tensor([0, 1]))


def test_read_data_uncalled4():
    config = Config(BASIC_CONFIG)

    worker = Uncalled4Worker(1, None, None, None, 0, 'cpu', config)
    worker.setup()
    ref_id = 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|'
    fasta_fh = Fasta(config.get_fasta_ref())
    ref_seq = str(fasta_fh[ref_id])

    transcript = Transcript(1, ref_id, ref_seq)

    data, samples, conditions = worker._read_data(transcript)

    assert data.shape == (2554, 12, 2)
    assert not np.isnan(data[301:303, :, :]).any()

    # Make sure that the reads are ordered.
    # This is necessary in order to guarantee
    # determinism of the GMM fitting (The order
    # of the data is important, because the k++
    # initialization selects random points to
    # initalize the GMM).
    expected_sample_ids = np.array([3, 3, 0, 1, 3, 2, 0, 1, 2, 1, 0, 2])
    assert np.array_equal(samples, expected_sample_ids)

    expected_condition_ids = np.array([1, 1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1])
    assert np.array_equal(conditions, expected_condition_ids)


def test_read_data_eventalign():
    config = Config(BASIC_EVENTALIGN_CONFIG)

    worker = GenericWorker(1, None, None, None, 0, 'cpu', config)
    worker.setup()
    ref_id = 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|'
    fasta_fh = Fasta(config.get_fasta_ref())
    ref_seq = str(fasta_fh[ref_id])

    transcript = Transcript(1, ref_id, ref_seq)

    data, samples, conditions = worker._read_data(transcript)
    print(data)
    print(samples, conditions)

    # In eventalign we have additional information
    # from the resquigglers that we can use to see
    # if a kmer is valid (we can check if there's
    # a mismatch between the model and reference kmer
    # or if the model kmer is NNNNN).
    # We then filter reads that have too many invalid
    # positions. Hence, here we have only 4 reads.
    assert data.shape == (2554, 4, 2)

    # Make sure that the samples are ordered.
    # This is necessary in order to guarantee
    # determinism of the GMM fitting (The order
    # of the data is important, because the k++
    # initialization selects random points to
    # initalize the GMM).
    # Note, the DB returns the reads in the same
    # order, so we can just order the samples
    # and we don't need to spend time ordering
    # by read or even read the read id.
    expected_sample_ids = np.array([0, 0, 1, 2])
    assert np.array_equal(samples, expected_sample_ids)

    expected_condition_ids = np.array([0, 0, 0, 1])
    assert np.array_equal(conditions, expected_condition_ids)


def test_get_transcripts_for_processing_insufficient_coverage():
    config = Config(BASIC_EVENTALIGN_CONFIG)
    run_cmd = RunCmd(config)

    refs = run_cmd._get_transcripts_for_processing()

    assert isinstance(refs, set)
    assert len(refs) == 0


def test_get_transcripts_for_processing_sufficient_coverage():
    config_yaml = copy.deepcopy(BASIC_CONFIG)
    config_yaml['min_coverage'] = 2
    config = Config(config_yaml)
    run_cmd = RunCmd(config)

    refs = run_cmd._get_transcripts_for_processing()
    refs = {ref.ref_id for ref in refs}

    assert isinstance(refs, set)
    assert len(refs) == 2
    assert refs == {'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|',
                    'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|'}


def test_get_transcripts_for_processing_selected_refs(tmpdir):
    p = tmpdir.join("selected_refs.txt")
    p.write("ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|\n")

    config_yaml = copy.deepcopy(BASIC_CONFIG)
    config_yaml['min_coverage'] = 2
    config_yaml['selected_refs'] = str(p)
    config = Config(config_yaml)
    run_cmd = RunCmd(config)

    refs = run_cmd._get_transcripts_for_processing()
    refs = {ref.ref_id for ref in refs}

    assert isinstance(refs, set)
    assert len(refs) == 1
    assert refs == {'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|'}


def test_get_transcripts_for_processing_ignored_refs(tmpdir):
    p = tmpdir.join("ignored_refs.txt")
    p.write("ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|\n")

    config_yaml = copy.deepcopy(BASIC_CONFIG)
    config_yaml['min_coverage'] = 2
    config_yaml['ignored_refs'] = str(p)
    config = Config(config_yaml)
    run_cmd = RunCmd(config)

    refs = run_cmd._get_transcripts_for_processing()
    refs = {ref.ref_id for ref in refs}

    assert isinstance(refs, set)
    assert len(refs) == 1
    assert refs == {'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|'}




import datetime
import inspect
import multiprocessing as mp
import os
import sys
import time

from collections import OrderedDict
from collections import namedtuple
from enum import Enum

import numpy as np
import pysam

from jaxtyping import Float
from loguru import logger
from pyfaidx import Fasta


Kit = Enum('Kit', ['RNA002', 'RNA004'])

# Center positions taken from the Uncalled4 paper
# https://doi.org/10.1101/2024.03.05.583511 (Fig. 1a
# for RNA002 and Supplementary Fig. 4 for RNA004).
# The center position represents the base that
# has the largest influence on the measured current
# levels for the kmer. It's defined with a one-based
# index.
Kit.RNA002.len = 5
Kit.RNA002.center = 4
Kit.RNA004.len = 9
Kit.RNA004.center = 5


READ_ID_TYPE = np.uint32
SAMPLE_ID_TYPE = np.uint8
MEASUREMENTS_TYPE = np.float32


INTENSITY_POS = 0
DWELL_POS = 1
MOTOR_DWELL_POS = 2

EVENTALIGN = 'eventalign'
UNCALLED4 = 'uncalled4'
REMORA = 'remora'

TranscriptRow = namedtuple('TranscriptRow', 'ref_id id')


def get_pos_kmer(pos, seq, kit):
    """
    Takes a position <pos> (in 0-based transcriptomic coords)
    and the reference sequence and returns the k-mer that
    starts at the given position with k-mer length depending
    on the kit.
    """
    return seq[pos:pos+kit.len]


class NanocomporeError(Exception):
    """ Basic exception class for nanocompore module """
    def __init__(self, message=""):
        super().__init__(message)


class NanocomporeWarning(Warning):
    """ Basic Warning class for nanocompore module """
    pass


def log_init_state(loc):
    logger.debug("\ttimestamp: {}".format(str(datetime.datetime.now())))
    for i, j in loc.items():
        if type(j) in [int, float, complex, list, dict, str, bool, set, tuple]: # Avoid non standard types
            logger.debug("\t{}: {}".format(i,j))


def mkdir(fn, exist_ok=False):
    """ Create directory recursivelly. Raise IO error if path exist or if error at creation """
    try:
        os.makedirs(fn, exist_ok=exist_ok)
    except Exception:
        raise NanocomporeError("Error creating output folder `{}`".format(fn))


def doc_func(func):
    """Parse the function description string"""

    docstr_list = []
    for l in inspect.getdoc(func).split("\n"):
        if l.startswith("*"):
            break
        else:
            docstr_list.append(l)

    return "\n".join(docstr_list)


def make_arg_dict(func):
    """Parse the arguments default value, type and doc"""

    # Init method for classes
    if inspect.isclass(func):
        func = func.__init__

    if inspect.isfunction(func) or inspect.ismethod(func):
        # Parse arguments default values and annotations
        d = OrderedDict()
        for name, p in inspect.signature(func).parameters.items():
            if p.name not in ["self","cls"]: # Object stuff. Does not make sense to include in doc
                d[name] = OrderedDict()
                if name not in ["kwargs","args"]: # Include but skip default required and type
                    # Get Annotation
                    if p.annotation != inspect._empty:
                        d[name]["type"] = p.annotation
                    # Get default value if available
                    if p.default == inspect._empty:
                        d[name]["required"] = True
                    else:
                        d[name]["default"] = p.default

        # Parse the docstring in a dict
        docstr_dict = OrderedDict()
        lab = None
        for l in inspect.getdoc(func).split("\n"):
            l = l.strip()
            if l:
                if l.startswith("*"):
                    lab = l[1:].strip()
                    docstr_dict[lab] = []
                elif lab:
                    docstr_dict[lab].append(l)

        # Concatenate and copy doc in main dict
        for name in d.keys():
            if name in docstr_dict:
                d[name]["help"] = " ".join(docstr_dict[name])
        return d


def arg_opt(func, arg, **kwargs):
    """Get options corresponding to argumant name and deal with special cases"""
    arg_dict = make_arg_dict(func)[arg]

    if "default" in arg_dict and "help" in arg_dict:
        arg_dict["help"] += " (default: %(default)s)"

    if "type" in arg_dict and "help" in arg_dict:
        arg_dict["help"] += " [%(type)s]"

    # Special case for boolean args
    if arg_dict["type"] is bool:
        if arg_dict["default"] is False:
            arg_dict["action"] = 'store_true'
            del arg_dict["type"]
        elif arg_dict["default"] is True:
            arg_dict["action"] = 'store_false'
            del arg_dict["type"]

    # Special case for lists args
    elif arg_dict["type"] is list:
        arg_dict["nargs"] = '*'

    return arg_dict


def jhelp(f:"python function or method"):
    """
    Display a Markdown pretty help message for functions and class methods (default __init__ is a class is passed)
    jhelp also display default values and type annotations if available.
    The docstring synthax should follow the same synthax as the one used for this function
    * f
        Function or method to display the help message for
    """
    # Private import as this is only needed if using jupyter
    from IPython.core.display import display, Markdown

    f_doc = doc_func(f)
    arg_doc = make_arg_dict(f)

    # Signature and function documentation
    s = "**{}** ({})\n\n{}\n\n---\n\n".format(f.__name__, ", ".join(arg_doc.keys()), f_doc)

    # Args doc
    for arg_name, arg_val in arg_doc.items():
        # Arg signature section
        s+= "* **{}**".format(arg_name)
        if "default" in arg_val:
            s+= " (default: {})".format(arg_val["default"])
        if "required" in arg_val:
            s+= " (required)"
        if "type" in arg_val:
            if type(list) == type:
                s+= " [{}]".format(arg_val["type"].__name__)
            else:
                s+= " [{}]".format(arg_val["type"])
        s+="\n\n"
        # Arg doc section
        if "help" in arg_val:
            s+= "{}\n\n".format(arg_val["help"])

    # Display in Jupyter
    display (Markdown(s))


def is_valid_fasta(file):
    """
    Returns a boolean indicating whether the given file is a valid FASTA file.
    """
    try:
        with Fasta(file) as fasta:
            # Fasta will return an empty iterator if the file is not properly
            # formatted FASTA file.
            return any(fasta)
    except IOError:
        # raise NanocomporeError("The fasta file cannot be opened")
        return False


def encode_kmer(kmer):
    encoding = 0
    for base in kmer:
        encoding <<= 2
        if base == "A" or base == "a":
            encoding |= 0
        elif base == "G" or base == "g":
            encoding |= 1
        elif base == "C" or base == "c":
            encoding |= 2
        elif base == "T" or base == "t":
            encoding |= 3
        else:
            raise RuntimeError(f"Bad nucleotide in kmer: {kmer}")
    return encoding


def decode_kmer(encoding, kmer_size):
    kmer = []
    for _ in range(kmer_size):
        base_code = encoding & ~(~0 << 2)
        base = ["A", "G", "C", "T"][base_code]
        kmer.append(base)
        encoding >>= 2
    return ''.join(kmer[::-1])


def monitor_workers(workers, delay_sec=5):
    """
    Check every *delay_sec* if any of the
    worker processes have been terminated
    with an error. If so, log an error and
    kill the all remaining workers and self.
    """
    terminated_ok = 0
    while terminated_ok < len(workers):
        time.sleep(delay_sec)
        for proc in workers:
            if proc.exitcode is not None:
                if proc.exitcode == 0:
                    terminated_ok += 1
                    continue
                logger.error(f"ERROR: A worker encountered an error (exitcode: {proc.exitcode}). "
                             "Will terminate all other workers and stop.")
                for child in mp.active_children():
                    child.terminate()
                sys.exit(1)


def get_references_from_bam(bam_path: str, has_data: bool=True) -> dict[str, int]:
    """
    Return references from the BAM's index.

    Parameters
    ----------
    bam_path : str
        Path to the BAM file.
    has_data : bool
        If set to true, return only references that have
        at least one mapped read to them.

    Returns
    -------
    dict[str, int]
        ref_id => number of reads
    """
    references = {}
    bam = pysam.AlignmentFile(bam_path, "rb")
    for line in bam.get_index_statistics():
        if (has_data and line.mapped > 0) or not has_data:
            references[line.contig] = line.mapped
    return references


def get_reads_invalid_ratio(intensity: Float[np.ndarray, "positions reads"]): 
    """
    Calculate the ratio of missing kmers
    in the read. It takes the kmers with
    min and max position to determine the
    read length.

    This is the method employed for Uncalled4 and Remora.
    For eventalign there's a custom code (in eventalign collapse),
    that takes in consideration the richer information provided
    by the resquiggler.

    Parameters
    ----------
    intensity : npt.NDArray[np.float32]
        2D array with shape (positions, reads).
        The missing values should be np.nan. 
    Returns
    -------
    np.array
        Array with shape (reads,) containing
        the ratio of invalid positions for
        each read.
    """
    nreads = intensity.shape[1]
    ratios = np.empty(nreads)
    for n in range(nreads):
        valid_positions = np.where(~np.isnan(intensity[:, n]))[0]
        length = valid_positions.max() - valid_positions.min() + 1
        invalid_ratio = (length - len(valid_positions))/length
        ratios[n] = invalid_ratio
    return ratios




"""
Parser for Uncalled4 resquiggling data.

Uncalled4 adds the resquiggling data as additional tags
in the BAM files. This parser reads the BAM file produced
by Uncalled4 and extracts the data from the tags.
"""

from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from typing import Dict, Union

import numpy as np
import pysam

from jaxtyping import Float, Int

from nanocompore.common import INTENSITY_POS
from nanocompore.common import DWELL_POS
from nanocompore.common import Kit


def get_reads(args):
    bam, ref_id, sample = args
    return [(read, sample) for read in bam.fetch(ref_id)]


class Uncalled4:
    def __init__(self,
                 ref_id: str,
                 ref_len: int,
                 bams: Dict[Union[str, int], pysam.AlignmentFile],
                 kit: Kit):
        """
        Initialize an Uncalled4 parser.

        Parameters
        ----------
        ref_id : str
            Reference id for the transcript that will be parsed.
        ref_len : int
            The length of the reference sequence.
        bams : Dict[Union[str, int], pysam.AlignmentFile]
            Dictionary that maps the sample label to the BAM file for the
            sample, opened as a pysam.AlignmentFile.
        kit : Kit
            The chemistry kit used for sequencing the samples.
        """
        self._ref_len = ref_len
        self._ref_id = ref_id
        self._bams = bams
        self._kit = kit


    def get_data(self) -> tuple[Float[np.ndarray, "positions reads vars"],
                                Int[np.ndarray, "reads"],
                                Int[np.ndarray, "reads"]]:
        """
        Returns the signal data for the transcript's reads.

        Returns
        -------
        tuple[Float[np.ndarray, "positions reads vars"],
              Int[np.ndarray, "reads"],
              Int[np.ndarray, "reads"]]
            Tuple with:
            - Tensor with shape (Positions, Reads, Vars)
              containing the signal measurements.
            - 1D array of size <Reads> with read ids.
            - 1D array of size <Reads> with sample labels.
        """
        read_ids = []
        sample_labels = []

        read_count = sum(bam.count(reference=self._ref_id)
                          for bam in self._bams.values())

        # shape is: (positions, reads, vars)
        reads_tensor = np.full((self._ref_len, read_count, 2), np.nan, dtype=np.float32)

        n_samples = len(self._bams)
        read_index = 0
        with ThreadPoolExecutor(max_workers=n_samples) as executor:
            futures = [executor.submit(get_reads, (bam, self._ref_id, sample))
                       for sample, bam in self._bams.items()]
            for future in as_completed(futures):
                sample_reads = future.result()
                for read, sample in sample_reads:
                    if read.is_secondary or read.is_supplementary:
                        continue

                    self._copy_signal_to_tensor(read, reads_tensor, read_index)

                    read_ids.append(read.query_name)
                    sample_labels.append(sample)

                    read_index += 1

        # Since we may have skipped secondary or supplementary reads,
        # we need to trim the tensor that we preallocated.
        reads_tensor = reads_tensor[:, :read_index, :]

        # Uncalled4 will use the minimum value in int16 (i.e. -32768)
        # to encode NaN values for the measurements. We don't want to
        # keep those as valid measurements, because they can lead to
        # false positives.
        reads_tensor = np.where(reads_tensor == -32768, np.nan, reads_tensor)

        if len(read_ids) == 0:
            return reads_tensor, np.array([]), np.array([])

        read_ids = np.array(read_ids)
        sample_labels = np.array(sample_labels)

        return reads_tensor, read_ids, sample_labels


    def _copy_signal_to_tensor(self, read, tensor, read_index):
        """
        Extracts the dwell, intensity and intensity std
        for a given read.
        """

        # ur gives the reference positions of the measurements.
        # ur is a list of start, end pairs of aligned segments, e.g:
        # start1, end1, start2, end2
        # ur uses 0-based indexing and [start, end) intervals,
        # but it includes the full spans of the evaluated kmers,
        # hence the number of measurements will be less than the
        # positions in the range. Specifically, we would have
        # N-K+1, where N is the reference positions and K is the
        # kmer size.
        ur = read.get_tag('ur')

        kit = self._kit

        # The signal measurements are ordered in
        # 3' to 5' direction so we invert them.
        # uc is the signal's current intensity
        uc = np.array(read.get_tag('uc')[::-1])
        # ud is the standard deviation in the current
        # We don't use the std for now
        # ud = np.array(read.get_tag('ud')[::-1])
        # ul is a list of dwell times for the kmers
        # The negative length values are paddings used
        # for each alignment segment.
        ul = np.array([d for d in read.get_tag('ul')[::-1] if d >= 0])
        ul = self._resolve_skips(ul)

        segments = [(ur[start_ind], ur[start_ind+1])
                    for start_ind in range(0, len(ur), 2)]

        signal_start = 0
        for ref_start, ref_end in segments:
            # Uncalled4 assigns the measurement to the
            # position of the base that had the highest
            # contribution (so called central position
            # of the kmer, which depending on the pore
            # may not be in the center).
            # The ranges in the ur tag however include
            # all positions that contributed to the
            # measurements. Hence, when we have N
            # measurements, the interval in ur will
            # cover N+K-1 positions (where K is the
            # kmer size for the given pore model).
            # We cut the K-1 positions from the first
            # and the last intervals.
            if ref_start == ur[0]:
                ref_start += kit.len - kit.center
            if ref_end == ur[-1]:
                ref_end -= kit.center - 1
            # Since we want to report the beginning of the
            # kmer instead of the central (most influential)
            # position we shift the signal.
            # Note that the magnitude of this shift is
            # the number of positions between the center and
            # the end of the kmer, because the RNA is read
            # in 3' to 5' direction, so the kmer is inverted
            # with respect to the 5'-3' sequence.
            ref_start -= kit.len - kit.center
            ref_end -= kit.len - kit.center

            segment_size = ref_end - ref_start
            signal_end = signal_start + segment_size

            # Copy signal values to the reads tensor
            tensor[ref_start:ref_end, read_index, DWELL_POS] = ul[signal_start:signal_end]
            tensor[ref_start:ref_end, read_index, INTENSITY_POS] = uc[signal_start:signal_end]
            # We don't use the std for now
            # tensor[read_index, ref_start:ref_end, 2] = ud[signal_start:signal_end]

            signal_start += segment_size


    def _resolve_skips(self, ul):
        resolved = np.empty(len(ul))
        resolved[0] = ul[0]
        for i in range(1, len(ul)):
            resolved[i] = ul[i] if ul[i] != 0 else resolved[i - 1]
        return resolved




import argparse
import math

from typing import Union

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np
import pandas as pd
import seaborn as sns
import torch

from gmm_gpu.gmm import GMM
from matplotlib.figure import Figure, SubFigure
from pyfaidx import Fasta

from nanocompore.api import get_pos, get_reads
from nanocompore.common import UNCALLED4
from nanocompore.common import INTENSITY_POS
from nanocompore.common import DWELL_POS
from nanocompore.common import decode_kmer
from nanocompore.comparisons import nanstd
from nanocompore.config import Config
from nanocompore.database import ResultsDB
from nanocompore.run import Uncalled4Worker, GenericWorker
from nanocompore.transcript import Transcript


TEST_REF = "ENST00000464651.1|ENSG00000166136.16|OTTHUMG00000019346.4|OTTHUMT00000051221.1|NDUFB8-204|NDUFB8|390|retained_intron|"

TEST_PVALUE_COLUMNS = {
    'GMM': ['GMM_chi2_pvalue'],
    'KS': ['KS_intensity_pvalue', 'KS_dwell_pvalue'],
    'TT': ['TT_intensity_pvalue', 'TT_dwell_pvalue'],
    'MW': ['MW_intensity_pvalue', 'MW_dwell_pvalue'],
    'AUTO': ['auto_pvalue'],
}

def plot_pvalues(config: Config,
                 reference: str,
                 start: Union[int, None]=None,
                 end: Union[int, None]=None,
                 kind: str='lineplot',
                 threshold: Union[float, None]=0.01,
                 figsize: tuple[int, int]=(30, 10),
                 tests: Union[list[str], None]=None,
                 palette: str='Dark2') -> Union[Figure, SubFigure, None]:
    """
    Plot the p-values from the statistical tests
    performed in a Nanocompore run.

    Parameters
    ----------
    config : Config
        The configuration object for the run.
    reference : str
        Transcript reference.
    start : Union[int, None]
        Start of the region that will be plotted.
    end : Union[int, None]
        End of the region that will be plotted.
    kind : str
        Kind of plot to make. The available options
        are: lineplot, barplot
    threshold : Union[float, None]
        If set, it will indicate the p-value threshold
        as a dashed horizontal line.
    figsize : tuple[int, int]
        Size of the figure.
    tests : Union[list[str], None]
        List of tests to plot. The available options
        are: GMM, KS, TT, MW.
        If set to None (default) it would plot all
        tests that were listed in the configuration.
    palette : str
        Color palette to use.

    Returns
    -------
    Union[Figure, SubFigure, None]
       The resulting figure that will contain the
       p-values for the specified tests in the
       provided region.
    """
    db = ResultsDB(config)
    cols = []
    if tests is None:
        tests = config.get_comparison_methods()
    for test in tests:
        if test not in TEST_PVALUE_COLUMNS:
            raise KeyError(f"Test {test} not supported.")
        for col in TEST_PVALUE_COLUMNS[test]:
            cols.append(col)
    pvals = db.get_columns_for_ref(cols + ['kmer'], reference)
    pvals = pd.melt(pvals,
                    id_vars=['pos', 'kmer'],
                    value_vars=[c for c in pvals.columns if c != 'pos'],
                    var_name='test',
                    value_name='pval')
    pvals['pval'] = -np.log10(pvals.pval)
    pvals['kmer'] = pvals.kmer.apply(lambda k: decode_kmer(k, config.get_kit().len))
    if start is not None and end is not None:
        pvals = pvals[pvals.pos.between(start, end)]
    elif start is not None:
        pvals = pvals[pvals.pos >= start]
    elif end is not None:
        pvals = pvals[pvals.pos <= end]

    max_pos = pvals.pos.max()
    min_pos = pvals.pos.min()
    is_detailed_plot = max_pos - min_pos <= 50

    fig, ax = plt.subplots(figsize=figsize)

    if threshold is not None:
        ax.axhline(y=-np.log10(threshold), color='grey', linestyle='--', label=f'p-value = {threshold}')

    if kind == 'lineplot' and is_detailed_plot:
        sns.pointplot(pvals, x='pos', y='pval', hue='test', ax=ax)
    elif kind == 'lineplot':
        sns.lineplot(pvals, x='pos', y='pval', hue='test', ax=ax)
    elif kind == 'barplot':
        sns.barplot(pvals, x='pos', y='pval', hue='test', ax=ax)
    else:
        raise ValueError(f'Plot kind {kind} not supported.')

    ax.set_xlabel("Reference position")
    ax.set_ylabel("-log10(pvalue)")

    if is_detailed_plot:
        ax2 = ax.twiny()
        ax2.spines['bottom'].set_position(('axes', -0.18))
        ax2.spines['bottom'].set_visible(False)
        kmer_df = pvals.loc[:, ['pos', 'kmer']].drop_duplicates()
        ax2.set_xlim(min_pos - 0.5, max_pos + 0.5)
        ax2.set_xticks(kmer_df.pos)
        ax2.set_xticklabels(kmer_df.kmer, rotation=60)


    # plt.legend()
    # print(ax2.get_legend_handles_labels())
    # plt.legend(loc="upper left", bbox_to_anchor=(1, 1))
    sns.move_legend(ax, "upper left", bbox_to_anchor=(1, 1))
    plt.tight_layout()

    return fig


def plot_signal(config: Config,
                reference: str,
                start: Union[int, None]=None,
                end: Union[int, None]=None,
                kind: str='violinplot',
                figsize: tuple[int, int]=(30, 10),
                split_samples: bool=False,
                markersize: int=2,
                palette: str='Dark2') -> Union[Figure, SubFigure, None]:
    """
    Plot the raw signal values (intensity and dwell time)
    for the given reference and region. Note that this
    will plot all reads from the input files without
    applying all the filtering and downsampling that
    Nanocompore does.

    Parameters
    ----------
    config : Config
        The configuration object for the run.
    reference : str
        Transcript reference.
    start : Union[int, None]
        Start of the region that will be plotted.
    end : Union[int, None]
        End of the region that will be plotted.
    kind : str
        Kind of plot to make. The available options
        are: lineplot, barplot
    figsize : tuple[int, int]
        Size of the figure.
    split_samples : bool
        If True, all samples would be plotted separately.
        By default it's False and the samples are grouped
        by condition.
    markersize : int
        Size of the points (only used for the swarmplot).
    palette : str
        Color palette to use.

    Returns
    -------
    Union[Figure, SubFigure, None]
       The resulting figure that will contain the
       intensity and log-dwell-time plots for the
       specified reference and region.
    """
    data, _, samples, conditions = get_reads(config, reference)
    positions = np.arange(data.shape[1])

    data = data[:, start:end, :]
    positions = positions[start:end]

    nreads = data.shape[0]
    group_var = samples if split_samples else conditions
    group_var_name = 'Sample' if split_samples else 'Condition'
    long_data = []
    for r in range(nreads):
        for p, pos in enumerate(positions):
            long_data.append((group_var[r],
                              pos,
                              data[r, p, INTENSITY_POS],
                              np.log10(data[r, p, DWELL_POS])))
    df = pd.DataFrame(long_data, columns=[group_var_name, 'pos', 'intensity', 'dwell'])

    fig, ax = plt.subplots(2, 1, figsize=figsize)

    if kind == 'violinplot':
        sns.violinplot(df, x='pos', y='intensity', hue=group_var_name, ax=ax[0], inner='quart', split=not split_samples)
        sns.violinplot(df, x='pos', y='dwell', hue=group_var_name, ax=ax[1], inner='quart', split=not split_samples, legend=False)
    elif kind == 'swarmplot':
        sns.swarmplot(df, x='pos', y='intensity', hue=group_var_name, ax=ax[0], size=markersize, dodge=True)
        sns.swarmplot(df, x='pos', y='dwell', hue=group_var_name, ax=ax[1], size=markersize, dodge=True, legend=False)
    elif kind == 'boxenplot':
        sns.boxenplot(df, x='pos', y='intensity', hue=group_var_name, ax=ax[0])
        sns.boxenplot(df, x='pos', y='dwell', hue=group_var_name, ax=ax[1], legend=False)
    else:
        raise ValueError(f'Plot kind {kind} not supported.')

    ax[0].set_xlabel(None)
    ax[0].set_ylabel("intensity")
    ax[1].set_xlabel("Reference position")
    ax[1].set_ylabel("log10(dwell time)")

    sns.move_legend(ax[0], "upper left", bbox_to_anchor=(1, 1), markerscale=math.ceil(10/markersize))

    plt.tight_layout()

    return fig


def plot_coverage(config: Config,
                  reference: str,
                  start: Union[int, None]=None,
                  end: Union[int, None]=None,
                  figsize: tuple[int, int]=(30, 10),
                  split_samples: bool=False,
                  palette: str='Dark2') -> Union[Figure, SubFigure, None]:
    """
    Plot the read coverage over a reference for all samples analysed.
    Note that this would plot the input coverage before applying
    any filtering that Nanocompore does before the comparison.

    Parameters
    ----------
    config : Config
        The configuration object for the run.
    reference : str
        Transcript reference.
    start : Union[int, None]
        Start of the region that will be plotted.
    end : Union[int, None]
        End of the region that will be plotted.
    figsize : tuple[int, int]
        Size of the figure.
    split_samples : bool
        If True, all samples would be plotted separately.
        By default it's False and the samples are grouped
        by condition.
    palette : str
        Color palette to use.

    Returns
    -------
    Union[Figure, SubFigure, None]
        Figure with the coverage plot.

    """
    data, reads, samples, conditions = get_reads(config, reference)
    positions = np.arange(data.shape[1])[start:end]
    data = data[:, start:end, 0]

    if split_samples:
        group_var = 'sample'
        groups = np.array(samples)
    else:
        group_var = 'condition'
        groups = np.array(conditions)

    covs = {}
    all_group_labels = np.unique(groups)
    for group in all_group_labels:
        covs[group] = np.sum(~np.isnan(data[groups == group, :]), axis=0)
    covs['pos'] = positions
    covs = pd.DataFrame(covs)
    covs = pd.melt(covs,
                   id_vars=['pos'],
                   value_vars=all_group_labels,
                   var_name=group_var,
                   value_name='cov')

    fig, ax = plt.subplots(figsize=figsize)

    sns.lineplot(covs, x='pos', y='cov', hue=group_var, palette=palette, ax=ax)
    ax.axhline(y=config.get_min_coverage(), linestyle='--', color='grey', label='Minimum coverage')

    plt.legend()
    sns.move_legend(ax, "upper left", bbox_to_anchor=(1, 1))
    plt.tight_layout()

    ax.set_xlabel('Reference position')
    ax.set_ylabel('Coverage')

    return fig


def plot_position(config: Config,
                  reference: str,
                  position: int,
                  figsize: tuple[int, int]=(10, 10),
                  point_size: int=20,
                  xlim: Union[tuple[Union[int, None], Union[int, None]], None]=(None, None),
                  ylim: Union[tuple[Union[int, None], Union[int, None]], None]=(None, None),
                  show_kde: bool=True,
                  kde_levels: int=10,
                  palette: Union[str, None]='Dark2') -> Union[Figure, SubFigure, None]:
    """
    Plot the signal data for a given position.

    Parameters
    ----------
    config : Config
        The configuration object for the run.
    reference : str
        Transcript reference.
    position : int
        0-based index on the reference indicating the position.
    figsize : tuple[int, int]
        Size of the figure.
    point_size : int
        Size of the points.
    xlim : Union[tuple[Union[int, None], Union[int, None]], None]
        Limits of the x-axis.
    ylim : Union[tuple[Union[int, None], Union[int, None]], None]
        Limits of the y-axis.
    kde : bool
        Whether to show the KDEs.
    kde_levels : int
        How many levels of the KDEs to show.
    palette : Union[str, None]
        Palette that will be used to determine the
        colors of both points and the GMMs. If set,
        it will override point_palette and gmm_palette.

    Returns
    -------
    Union[Figure, SubFigure, None]
       The resulting figure that will contain a
       2D plot with the observations as points
       and the gaussians obtained from the GMM.
    """
    fig, ax = plt.subplots(figsize=figsize)
    if xlim is not None:
        ax.set_xlim(*xlim)
    if ylim is not None:
        ax.set_ylim(*ylim)
    df = get_pos(config, reference, position)
    sns.scatterplot(df,
                    x='dwell',
                    y='intensity',
                    hue='condition',
                    s=point_size,
                    palette=palette,
                    ax=ax)
    if show_kde:
        sns.kdeplot(df,
                    x='dwell',
                    y='intensity',
                    levels=kde_levels,
                    hue="condition",
                    palette=palette,
                    ax=ax)

    ax.set_xlabel('Dwell time')
    ax.set_ylabel('Intensity')

    return fig


def plot_gmm(config: Config,
             reference: str,
             position: int,
             figsize: tuple[int, int]=(10, 10),
             point_size: int=20,
             xlim: Union[tuple[Union[int, None], Union[int, None]], None]=(None, None),
             ylim: Union[tuple[Union[int, None], Union[int, None]], None]=(None, None),
             gmm_levels: int=4,
             palette: Union[str, None]='Dark2',
             point_palette: Union[str, None]=None,
             gmm_palette: Union[str, None]=None) -> Union[Figure, SubFigure, None]:
    """
    Plot the GMM fitted by Nanocompore for a given position.

    Parameters
    ----------
    config : Config
        The configuration object for the run.
    reference : str
        Transcript reference.
    position : int
        0-based index on the reference indicating the position.
    figsize : tuple[int, int]
        Size of the figure.
    point_size : int
        Size of the points.
    xlim : Union[tuple[Union[int, None], Union[int, None]], None]
        Limits of the x-axis.
    ylim : Union[tuple[Union[int, None], Union[int, None]], None]
        Limits of the y-axis.
    gmm_levels : int
        How many levels of the GMM to show.
    palette : Union[str, None]
        Palette that will be used to determine the
        colors of both points and the GMMs. If set,
        it will override point_palette and gmm_palette.
    point_palette : Union[str, None]
        Palette to use for the points.
    gmm_palette : Union[str, None]
        Palette to use for the GMMs.

    Returns
    -------
    Union[Figure, SubFigure, None]
       The resulting figure that will contain a
       2D plot with the observations as points
       and the gaussians obtained from the GMM.
    """
    fasta_fh = Fasta(config.get_fasta_ref())
    ref_seq = str(fasta_fh[reference])

    transcript = Transcript(1, reference, ref_seq)

    if config.get_resquiggler() == UNCALLED4:
        worker_class = Uncalled4Worker
    else:
        worker_class = GenericWorker
    worker = worker_class(1, None, None, None, None, 'cpu', config)
    data, samples, conditions = worker._read_data(transcript)
    prepared_data = worker._prepare_data(data, samples, conditions)
    data, samples, conditions, _ = prepared_data
    data = data[[position], :, :]
    max_reads = worker._conf.get_downsample_high_coverage()
    data, samples, conditions = worker._downsample(
            data, samples, conditions, max_reads)

    std = nanstd(data, 1).unsqueeze(1)
    outliers = (((data - data.nanmean(1, keepdim=True)) / std).abs() > 3).any(2)
    data[outliers] = np.nan

    # Standardize the data
    std = nanstd(data, 1)
    data = (data - data.nanmean(1).unsqueeze(1)) / std.unsqueeze(1)

    gmm = GMM(n_components=2,
              device='cpu',
              random_seed=42,
              dtype=torch.float32)
    gmm.fit(data)

    # Means is a list with the means for each component.
    # The shape of each is (Points, Dims). We have a single point.
    c1_mean = gmm.means[0][0]
    c2_mean = gmm.means[1][0]

    # Covs is a list with the cov matrices for the components.
    # The shape of each is (Points, Dims, Dims).
    c1_cov = gmm.covs[0][0]
    c2_cov = gmm.covs[1][0]

    x1, y1 = np.random.multivariate_normal(c1_mean, c1_cov, 1000).T
    x2, y2 = np.random.multivariate_normal(c2_mean, c2_cov, 1000).T
    sampled_gaussians = pd.DataFrame(
        {'x': np.concatenate([x1, x2]),
         'y': np.concatenate([y1, y2]),
         'cluster': np.concatenate([np.full((1000,), 'Cluster 1'),
                                    np.full((1000,), 'Cluster 2')])})

    df = pd.DataFrame(data[0], columns=['intensity', 'dwell'])
    cond_labels = config.get_condition_labels()
    df['condition'] = [cond_labels[c] for c in conditions]

    if palette is not None:
        palette = {cond_labels[0]: plt.get_cmap(palette).colors[0],
                   cond_labels[1]: plt.get_cmap(palette).colors[1],
                   'Cluster 1': plt.get_cmap(palette).colors[2],
                   'Cluster 2': plt.get_cmap(palette).colors[3]}
    else:
        palette = {cond_labels[0]: plt.get_cmap(point_palette).colors[0],
                   cond_labels[1]: plt.get_cmap(point_palette).colors[1],
                   'Cluster 1': plt.get_cmap(gmm_palette).colors[0],
                   'Cluster 2': plt.get_cmap(gmm_palette).colors[1]}

    fig, ax = plt.subplots(figsize=figsize)
    if xlim is not None:
        ax.set_xlim(*xlim)
    if ylim is not None:
        ax.set_ylim(*ylim)
    if xlim is None and ylim is None:
        lims = (min(df.dwell.min(),
                    df.intensity.min(),
                    sampled_gaussians.x.min(),
                    sampled_gaussians.y.min()),
                max(df.dwell.max(),
                    df.intensity.max(),
                    sampled_gaussians.x.max(),
                    sampled_gaussians.y.max()))
        ax.set_xlim(*lims)
        ax.set_ylim(*lims)

    sns.kdeplot(sampled_gaussians,
                x='x',
                y='y',
                levels=gmm_levels,
                hue="cluster",
                palette=palette,
                ax=ax)
    sns.scatterplot(df,
                    x='dwell',
                    y='intensity',
                    hue='condition',
                    s=point_size,
                    palette=palette,
                    legend=False,
                    ax=ax)

    handles = [mpatches.Patch(facecolor=col, label=label)
               for label, col in palette.items()]
    plt.legend(handles=handles)

    ax.set_xlabel('Standardized log10(dwell)')
    ax.set_ylabel('Standardized intensity')
    return fig




import os

import numpy as np
import pandas as pd
import polars as pl

from loguru import logger
from statsmodels.stats.multitest import multipletests
from gtfparse import read_gtf

from nanocompore.common import decode_kmer
from nanocompore.database import ResultsDB


class Postprocessor():
    def __init__ (self, config):
        self._outpath = config.get_outpath()
        self._prefix = config.get_outprefix()
        self._result_exists_strategy = config.get_result_exists_strategy()
        self._correction_method = config.get_correction_method()
        self._config = config
        self._db = ResultsDB(config, init_db=False)


    def __call__(self):
        logger.info("Creating database indices.")
        self._db.index_database()

        all_columns = self._db.get_result_column_names()

        # The test_result_columns are all columns from the database
        # that need to be added to the result TSV.
        test_result_columns = self._get_test_result_columns(all_columns)
        self._export_results_tsv(test_result_columns)

        if self._config.get_export_shift_stats():
            logger.info("Writing the shift stats to a file.")
            shift_stats_columns = self._get_shift_stats_columns()
            self._export_shift_stats(shift_stats_columns)
        else:
            logger.info("Exporting shift statistics to a TSV is disabled. "
                        "However, you can still find them in the result database.")


    def _export_results_tsv(self, test_result_columns, chunksize=1_000_000):
        out_tsv = os.path.join(self._outpath, f"{self._prefix}nanocompore_results.tsv")
        kmer_len = self._config.get_kit().len

        map_genomic_positions = bool(self._config.get_gtf())
        if map_genomic_positions:
            transcript_positions = self._get_transcript_positions(self._config.get_gtf())
            genomic_positions = np.vectorize(transcript_positions.get)

        # In order to do the multiple test correction we need
        # all p-values. However, we'd like to avoid loading
        # all data in memory. Hence, we read only the p-value
        # columns from the database and then the rest of the
        # results are processed in batches. I.e. we read each
        # batch, add the corresponding p-values and q-values
        # and write the append the batch to the results TSV.

        # We read the p-value columns we in memory and get add
        # the corrected q-values to them. We also get the
        # auto_test column if auto test was used, because we
        # need it in order to do the multiple test correction
        # independently for the different tests.
        pval_columns = [c for c in test_result_columns if 'pvalue' in c]
        p_values = self._db.get_columns(pval_columns)
        auto_test = None
        if 'auto_pvalue' in test_result_columns:
            auto_test = self._db.get_columns(['auto_test'])['auto_test']
        pq_values = self._correct_pvalues(p_values,
                                          method=self._correction_method,
                                          auto_test=auto_test)

        # The remaining columns will be loaded in batches from the db and written
        # to the output TSV.
        non_pqval_cols = [c for c in test_result_columns if 'pvalue' not in c]

        renamings = {'name': 'ref_id', 'kmer': 'ref_kmer'}

        base_columns = ['pos', 'name', 'kmer']

        # Since we have the p-values in memory,
        # we only want to get the base columns and
        # the non-p-value test result columns.
        select_columns = base_columns + non_pqval_cols
        test_columns = list(pq_values.columns) + non_pqval_cols

        output_columns = ['pos', 'chr', 'genomicPos', 'ref_id', 'strand', 'ref_kmer'] + test_columns

        logger.info(f"Writing the test results to: {out_tsv}")

        completed_rows = 0
        for chunk in self._db.get_results(select_columns, chunksize=chunksize):
            nrows = chunk.shape[0]

            chunk['genomicPos'] = 'NA'
            chunk['chr'] = 'NA'
            chunk['strand'] = 'NA'

            # Make sure counts are integer
            chunk = chunk.astype({col: 'int'
                                  for col in chunk.columns
                                  if col.endswith('_mod') or col.endswith('_unmod')})

            chunk['kmer'] = [decode_kmer(code, kmer_len) for code in chunk['kmer']]

            chunk.rename(columns=renamings, inplace=True)
            qvals_chunk = pq_values[completed_rows:completed_rows+nrows].reset_index(drop=True)
            chunk = pd.concat([chunk, qvals_chunk], axis=1)

            if map_genomic_positions:
                genomic_data = genomic_positions(chunk.ref_id.str.split('|').str[0])
                coords_getter = np.vectorize(
                        lambda tx, pos: (tx['seqname'],
                                         tx['strand'],
                                         tx['exons'](pos)))
                chrom, strand, gpos = coords_getter(genomic_data, chunk.pos)
                chunk['chr'] = chrom
                chunk['strand'] = strand
                chunk['genomicPos'] = gpos

            mode = 'w' if completed_rows == 0 else 'a'
            header = completed_rows == 0
            chunk.to_csv(out_tsv, sep='\t', columns=output_columns, index=False, header=header, mode=mode)

            completed_rows += nrows


    def _export_shift_stats(self, shift_stats_columns, chunksize=1_000_000):
        shift_stats_tsv = os.path.join(self._outpath, f"{self._prefix}nanocompore_shift_stats.tsv")
        base_columns = ['name', 'pos']
        renamings = {'name': 'ref_id'}
        select_columns = base_columns + shift_stats_columns
        for i, chunk in enumerate(self._db.get_results(select_columns, chunksize=chunksize)):
            chunk.rename(columns=renamings, inplace=True)
            chunk = chunk.round(decimals=4)
            mode = 'w' if i == 0 else 'a'
            header = i == 0
            chunk.to_csv(shift_stats_tsv, sep='\t', index=False, header=header, mode=mode)


    def _get_transcript_positions(self, gtf_path):
        gtf = read_gtf(gtf_path, features=['exon'])
        gtf = gtf.select('transcript_id',
                         'seqname',
                         'strand',
                         pl.struct('start', 'end').alias('exons'))\
                 .group_by('transcript_id', 'seqname', 'strand')\
                 .agg('exons')\
                 .to_pandas()\
                 .set_index('transcript_id')\
                 .to_dict('index')
        for transcript in gtf.values():
            transcript['exons'] = self._get_genomic_mapper(transcript['exons'],
                                                           transcript['strand'])
        return gtf


    def _get_genomic_mapper(self, exons, strand):
        if strand == '+':
            positions = np.concatenate([np.arange(e['start'], e['end']+1) for e in exons])
        else:
            positions = np.concatenate([np.arange(e['end'], e['start']-1, -1) for e in exons])
        # GTF files use 1-based indexing.
        # Hence, we subtract one to convert to 0-based indices.
        positions -= 1
        return np.vectorize(dict(zip(range(len(positions)),
                                     positions)).get)


    def _sort_headers_list(self, headers):
        return sorted(headers,
                      key=lambda x: (not ('pvalue' in x or 'qvalue' in x),
                                     ('_mod' in x or '_unmod' in x),
                                     x))


    def _get_test_result_columns(self, columns):
        keywords = ['pvalue', 'qvalue', 'LOR', 'GMM', 'logit', 'auto_test', '_mod', '_unmod']
        selected =  [col
                     for col in columns
                     if any(kw in col for kw in keywords)]
        return self._sort_headers_list(selected)


    def _get_shift_stats_columns(self):
        conditions = ['c1', 'c2']
        stats = ['mean', 'median', 'sd']
        values = ['intensity', 'dwell']
        return [f"{c}_{s}_{v}"
                for v in values
                for s in stats
                for c in conditions]


    def _correct_pvalues(self, df, method='fdr_bh', auto_test=None):
        for column in df.columns:
            if column == 'auto_pvalue' and method == 'fdr_bh':
                pvals = np.array(df[column].values, dtype=float)
                all_qvals = np.full(pvals.shape, np.nan, dtype=float)

                logger.debug(f"Starting to correct pvalues for {column} with {method}")
                # We need to correct the p-values for a total of N
                # tests. However, those p-values were obtained by
                # performing different tests and are not directly
                # comparable. Using the standard Benjamini-Hochberg
                # will introduce bias against tests that are less
                # sensitive by pushing their p-values to the bottom
                # of the ranking and then applying more severe penalties
                # for them. As a simple workaround, we perform the
                # correction separately (thus ordering different
                # tests independently) and then we multiply by
                # a correction coefficient to compensate for the
                # fact that the correction was done for less than
                # N tests.
                for test_type in auto_test.unique():
                    test_mask = auto_test == test_type
                    # We'll ignore NaNs when performing the correction.
                    qvals = self._multipletests_filter_nan(pvals[test_mask], method)
                    if len(qvals) > 0:
                        qvals *= len(pvals) / len(qvals)
                    all_qvals[test_mask] = qvals
                df['auto_qvalue'] = all_qvals
                logger.debug(f"pvalues for {column} have been corrected using {method}")
            elif 'pvalue' in column:
                pvals = np.array(df[column].values, dtype=float)

                # We'll ignore NaNs when performing the correction.
                logger.debug(f"Starting to correct pvalues for {column} with {method}")
                qvals = self._multipletests_filter_nan(pvals, method)

                # Replace the original p-values with the corrected p-values in the dataframe
                corrected_column = column.replace("pvalue", "qvalue")
                df[corrected_column] = qvals

                logger.debug(f"pvalues for {column} have been corrected using {method}")

        df.sort_index(axis=1, inplace=True)
        return df


    @staticmethod
    def _multipletests_filter_nan(pvalues: np.ndarray, method="fdr_bh") -> np.ndarray:
        """
        Performs p-value correction for multiple hypothesis testing
        using the method specified. The pvalues list can contain
        np.nan values, which are ignored during p-value correction.

        Parameters
        ----------
        pvalues : np.ndarray
            Array of p-values to correct.
        method :
            Which correction strategy to use. Should be supported
            by statsmodels's multipletests method.

        Returns
        -------
        np.ndarray
            Array of q-values.

        Example
        -------
        input = np.array([0.1, 0.01, np.nan, 0.01, 0.5, 0.4, 0.01, 0.001, np.nan, np.nan, 0.01, np.nan])
        Postprocessor._multipletests_filter_nan(input)
        array([0.13333333, 0.016, nan, 0.016, 0.5, 0.45714286, 0.016, 0.008, nan, nan, 0.016, nan])
        """
        nans = np.isnan(pvalues)
        corrected_p_values = multipletests(pvalues[~nans], method=method)[1]
        qvalues = np.copy(pvalues)
        qvalues[~nans] = corrected_p_values
        return qvalues

